import functools
import inspect
import json
import os
import traceback
from collections.abc import Callable
from typing import Any

from openai import APIError
from pydantic_ai import Agent, Tool
from pydantic_ai.mcp import MCPServer
from pydantic_ai.messages import (
    FinalResultEvent,
    FunctionToolCallEvent,
    FunctionToolResultEvent,
    ModelMessagesTypeAdapter,
    PartDeltaEvent,
    PartStartEvent,
    TextPartDelta,
    ToolCallPartDelta,
)
from pydantic_ai.models import Model
from pydantic_ai.settings import ModelSettings

from zrb.attr.type import StrAttr, fstring
from zrb.context.any_context import AnyContext
from zrb.context.any_shared_context import AnySharedContext
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.llm_config import LLMConfig
from zrb.llm_config import llm_config as default_llm_config
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.util.attr import get_attr, get_str_attr
from zrb.util.cli.style import stylize_faint
from zrb.util.file import read_file, write_file
from zrb.util.run import run_async

ListOfDict = list[dict[str, Any]]
ToolOrCallable = Tool | Callable


class LLMTask(BaseTask):
    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput | None] | AnyInput | None = None,
        env: list[AnyEnv | None] | AnyEnv | None = None,
        model: (
            Callable[[AnySharedContext], Model | str | fstring] | Model | None
        ) = None,
        render_model: bool = True,
        model_base_url: StrAttr | None = None,
        render_model_base_url: bool = True,
        model_api_key: StrAttr | None = None,
        render_model_api_key: bool = True,
        model_settings: (
            ModelSettings | Callable[[AnySharedContext], ModelSettings] | None
        ) = None,
        agent: Agent | Callable[[AnySharedContext], Agent] | None = None,
        system_prompt: StrAttr | None = None,
        render_system_prompt: bool = True,
        message: StrAttr | None = None,
        tools: (
            list[ToolOrCallable] | Callable[[AnySharedContext], list[ToolOrCallable]]
        ) = [],
        mcp_servers: (
            list[MCPServer] | Callable[[AnySharedContext], list[MCPServer]]
        ) = [],
        conversation_history: (
            ListOfDict | Callable[[AnySharedContext], ListOfDict]
        ) = [],
        conversation_history_reader: (
            Callable[[AnySharedContext], ListOfDict] | None
        ) = None,
        conversation_history_writer: (
            Callable[[AnySharedContext, ListOfDict], None] | None
        ) = None,
        conversation_history_file: StrAttr | None = None,
        render_history_file: bool = True,
        execute_condition: bool | str | Callable[[AnySharedContext], bool] = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        max_call_iteration: int = 20,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
        successor: list[AnyTask] | AnyTask | None = None,
    ):
        super().__init__(
            name=name,
            color=color,
            icon=icon,
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            execute_condition=execute_condition,
            retries=retries,
            retry_period=retry_period,
            readiness_check=readiness_check,
            readiness_check_delay=readiness_check_delay,
            readiness_check_period=readiness_check_period,
            readiness_failure_threshold=readiness_failure_threshold,
            readiness_timeout=readiness_timeout,
            monitor_readiness=monitor_readiness,
            upstream=upstream,
            fallback=fallback,
            successor=successor,
        )
        self._model = model
        self._render_model = render_model
        self._model_base_url = model_base_url
        self._render_model_base_url = render_model_base_url
        self._model_api_key = model_api_key
        self._render_model_api_key = render_model_api_key
        self._model_settings = model_settings
        self._agent = agent
        self._system_prompt = system_prompt
        self._render_system_prompt = render_system_prompt
        self._message = message
        self._tools = tools
        self._additional_tools: list[ToolOrCallable] = []
        self._mcp_servers = mcp_servers
        self._additional_mcp_servers: list[MCPServer] = []
        self._conversation_history = conversation_history
        self._conversation_history_reader = conversation_history_reader
        self._conversation_history_writer = conversation_history_writer
        self._conversation_history_file = conversation_history_file
        self._render_history_file = render_history_file
        self._max_call_iteration = max_call_iteration

    def add_tool(self, tool: ToolOrCallable):
        self._additional_tools.append(tool)

    def add_mcp_server(self, mcp_server: MCPServer):
        self._additional_mcp_servers.append(mcp_server)

    async def _exec_action(self, ctx: AnyContext) -> Any:
        history = await self._read_conversation_history(ctx)
        user_prompt = self._get_message(ctx)
        agent = self._get_agent(ctx)
        try:
            async with agent.run_mcp_servers():
                async with agent.iter(
                    user_prompt=user_prompt,
                    message_history=ModelMessagesTypeAdapter.validate_python(history),
                ) as agent_run:
                    async for node in agent_run:
                        # Each node represents a step in the agent's execution
                        # Reference: https://ai.pydantic.dev/agents/#streaming
                        try:
                            await self._print_node(ctx, agent_run, node)
                        except APIError as e:
                            # Extract detailed error information from the response
                            error_details = _extract_api_error_details(e)
                            ctx.log_error(f"API Error: {error_details}")
                            raise
                        except Exception as e:
                            ctx.log_error(f"Error processing node: {str(e)}")
                            ctx.log_error(f"Error type: {type(e).__name__}")
                            raise
                    new_history = json.loads(agent_run.result.all_messages_json())
                    await self._write_conversation_history(ctx, new_history)
                    return agent_run.result.data
        except Exception as e:
            ctx.log_error(f"Error in agent execution: {str(e)}")
            raise

    async def _print_node(self, ctx: AnyContext, agent_run: Any, node: Any):
        if Agent.is_user_prompt_node(node):
            # A user prompt node => The user has provided input
            ctx.print(stylize_faint(f">> UserPromptNode: {node.user_prompt}"))
        elif Agent.is_model_request_node(node):
            # A model request node => We can stream tokens from the model"s request
            ctx.print(
                stylize_faint(">> ModelRequestNode: streaming partial request tokens")
            )
            async with node.stream(agent_run.ctx) as request_stream:
                is_streaming = False
                async for event in request_stream:
                    if isinstance(event, PartStartEvent):
                        if is_streaming:
                            ctx.print("", plain=True)
                        ctx.print(
                            stylize_faint(
                                f"[Request] Starting part {event.index}: {event.part!r}"
                            ),
                        )
                        is_streaming = False
                    elif isinstance(event, PartDeltaEvent):
                        if isinstance(event.delta, TextPartDelta):
                            ctx.print(
                                stylize_faint(f"{event.delta.content_delta}"),
                                end="",
                                plain=is_streaming,
                            )
                        elif isinstance(event.delta, ToolCallPartDelta):
                            ctx.print(
                                stylize_faint(f"{event.delta.args_delta}"),
                                end="",
                                plain=is_streaming,
                            )
                        is_streaming = True
                    elif isinstance(event, FinalResultEvent):
                        if is_streaming:
                            ctx.print("", plain=True)
                        ctx.print(
                            stylize_faint(f"[Result] tool_name={event.tool_name}"),
                        )
                        is_streaming = False
                if is_streaming:
                    ctx.print("", plain=True)
        elif Agent.is_call_tools_node(node):
            # A handle-response node => The model returned some data, potentially calls a tool
            ctx.print(
                stylize_faint(
                    ">> CallToolsNode: streaming partial response & tool usage"
                )
            )
            async with node.stream(agent_run.ctx) as handle_stream:
                async for event in handle_stream:
                    if isinstance(event, FunctionToolCallEvent):
                        # Handle empty arguments across different providers
                        if event.part.args == "" or event.part.args is None:
                            event.part.args = {}
                        elif isinstance(
                            event.part.args, str
                        ) and event.part.args.strip() in ["null", "{}"]:
                            # Some providers might send "null" or "{}" as a string
                            event.part.args = {}
                        # Handle dummy property if present (from our schema sanitization)
                        if (
                            isinstance(event.part.args, dict)
                            and "_dummy" in event.part.args
                        ):
                            del event.part.args["_dummy"]
                        ctx.print(
                            stylize_faint(
                                f"[Tools] The LLM calls tool={event.part.tool_name!r} with args={event.part.args} (tool_call_id={event.part.tool_call_id!r})"  # noqa
                            )
                        )
                    elif isinstance(event, FunctionToolResultEvent):
                        ctx.print(
                            stylize_faint(
                                f"[Tools] Tool call {event.tool_call_id!r} returned => {event.result.content}"  # noqa
                            )
                        )
        elif Agent.is_end_node(node):
            # Once an End node is reached, the agent run is complete
            ctx.print(stylize_faint(f"{agent_run.result.data}"))

    async def _write_conversation_history(
        self, ctx: AnyContext, conversations: list[Any]
    ):
        if self._conversation_history_writer is not None:
            await run_async(self._conversation_history_writer(ctx, conversations))
        history_file = self._get_history_file(ctx)
        if history_file != "":
            write_file(history_file, json.dumps(conversations, indent=2))

    def _get_model_settings(self, ctx: AnyContext) -> ModelSettings | None:
        if callable(self._model_settings):
            return self._model_settings(ctx)
        return self._model_settings

    def _get_agent(self, ctx: AnyContext) -> Agent:
        if isinstance(self._agent, Agent):
            return self._agent
        if callable(self._agent):
            return self._agent(ctx)
        tools_or_callables = list(
            self._tools(ctx) if callable(self._tools) else self._tools
        )
        tools_or_callables.extend(self._additional_tools)
        tools = [
            tool if isinstance(tool, Tool) else Tool(_wrap_tool(tool), takes_ctx=False)
            for tool in tools_or_callables
        ]
        mcp_servers = list(
            self._mcp_servers(ctx) if callable(self._mcp_servers) else self._mcp_servers
        )
        mcp_servers.extend(self._additional_mcp_servers)
        return Agent(
            self._get_model(ctx),
            system_prompt=self._get_system_prompt(ctx),
            tools=tools,
            mcp_servers=mcp_servers,
            model_settings=self._get_model_settings(ctx),
            retries=3,
        )

    def _get_model(self, ctx: AnyContext) -> str | Model | None:
        model = get_attr(ctx, self._model, None, auto_render=self._render_model)
        if model is None:
            return default_llm_config.get_default_model()
        if isinstance(model, str):
            model_base_url = self._get_model_base_url(ctx)
            model_api_key = self._get_model_api_key(ctx)
            llm_config = LLMConfig(
                default_model_name=model,
                default_base_url=model_base_url,
                default_api_key=model_api_key,
            )
            if model_base_url is None and model_api_key is None:
                default_model_provider = default_llm_config.get_default_model_provider()
                if default_model_provider is not None:
                    llm_config.set_default_provider(default_model_provider)
            return llm_config.get_default_model()
        raise ValueError(f"Invalid model: {model}")

    def _get_model_base_url(self, ctx: AnyContext) -> str | None:
        base_url = get_attr(
            ctx, self._model_base_url, None, auto_render=self._render_model_base_url
        )
        if isinstance(base_url, str) or base_url is None:
            return base_url
        raise ValueError(f"Invalid model base URL: {base_url}")

    def _get_model_api_key(self, ctx: AnyContext) -> str | None:
        api_key = get_attr(
            ctx, self._model_api_key, None, auto_render=self._render_model_api_key
        )
        if isinstance(api_key, str) or api_key is None:
            return api_key
        raise ValueError(f"Invalid model API key: {api_key}")

    def _get_system_prompt(self, ctx: AnyContext) -> str:
        system_prompt = get_attr(
            ctx,
            self._system_prompt,
            None,
            auto_render=self._render_system_prompt,
        )
        if system_prompt is not None:
            return system_prompt
        return default_llm_config.get_default_system_prompt()

    def _get_message(self, ctx: AnyContext) -> str:
        return get_str_attr(ctx, self._message, "How are you?", auto_render=True)

    async def _read_conversation_history(self, ctx: AnyContext) -> ListOfDict:
        if self._conversation_history_reader is not None:
            return await run_async(self._conversation_history_reader(ctx))
        if callable(self._conversation_history):
            return self._conversation_history(ctx)
        history_file = self._get_history_file(ctx)
        if (
            len(self._conversation_history) == 0
            and history_file != ""
            and os.path.isfile(history_file)
        ):
            return json.loads(read_file(history_file))
        return self._conversation_history

    def _get_history_file(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx,
            self._conversation_history_file,
            "",
            auto_render=self._render_history_file,
        )


def _wrap_tool(func):
    sig = inspect.signature(func)
    if len(sig.parameters) == 0:

        @functools.wraps(func)
        async def wrapper(_dummy=None):
            try:
                return await run_async(func())
            except Exception as e:
                # Optionally, you can include more details from traceback if needed.
                error_details = traceback.format_exc()
                return json.dumps({"error": f"{e}", "details": f"{error_details}"})

        new_sig = inspect.Signature(
            parameters=[
                inspect.Parameter(
                    "_dummy", inspect.Parameter.POSITIONAL_OR_KEYWORD, default=None
                )
            ]
        )
        # Override the wrapper's signature so introspection yields a non-empty schema.
        wrapper.__signature__ = new_sig
        return wrapper
    else:

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                return await run_async(func(*args, **kwargs))
            except Exception as e:
                # Optionally, you can include more details from traceback if needed.
                error_details = traceback.format_exc()
                return json.dumps({"error": f"{e}", "details": f"{error_details}"})

        return wrapper


def _extract_api_error_details(error: APIError) -> str:
    """Extract detailed error information from an APIError."""
    details = f"{error.message}"
    # Try to parse the error body as JSON
    if error.body:
        try:
            if isinstance(error.body, str):
                body_json = json.loads(error.body)
            elif isinstance(error.body, bytes):
                body_json = json.loads(error.body.decode("utf-8"))
            else:
                body_json = error.body
            # Extract error details from the JSON structure
            if isinstance(body_json, dict):
                if "error" in body_json:
                    error_obj = body_json["error"]
                    if isinstance(error_obj, dict):
                        if "message" in error_obj:
                            details += f"\nProvider message: {error_obj['message']}"
                        if "code" in error_obj:
                            details += f"\nError code: {error_obj['code']}"
                        if "status" in error_obj:
                            details += f"\nStatus: {error_obj['status']}"
                # Check for metadata that might contain provider-specific information
                if "metadata" in body_json and isinstance(body_json["metadata"], dict):
                    metadata = body_json["metadata"]
                    if "provider_name" in metadata:
                        details += f"\nProvider: {metadata['provider_name']}"
                    if "raw" in metadata:
                        try:
                            raw_json = json.loads(metadata["raw"])
                            if "error" in raw_json and isinstance(
                                raw_json["error"], dict
                            ):
                                raw_error = raw_json["error"]
                                if "message" in raw_error:
                                    details += (
                                        f"\nRaw error message: {raw_error['message']}"
                                    )
                        except (KeyError, TypeError, ValueError):
                            # If we can't parse the raw JSON, just include it as is
                            details += f"\nRaw error data: {metadata['raw']}"
        except json.JSONDecodeError:
            # If we can't parse the JSON, include the raw body
            details += f"\nRaw error body: {error.body}"
        except Exception as e:
            # Catch any other exceptions during parsing
            details += f"\nError parsing error body: {str(e)}"
    # Include request information if available
    if hasattr(error, "request") and error.request:
        if hasattr(error.request, "method") and hasattr(error.request, "url"):
            details += f"\nRequest: {error.request.method} {error.request.url}"
        # Include a truncated version of the request content if available
        if hasattr(error.request, "content") and error.request.content:
            content = error.request.content
            if isinstance(content, bytes):
                try:
                    content = content.decode("utf-8")
                except UnicodeDecodeError:
                    content = str(content)
            details += f"\nRequest content: {content}"
    return details



import importlib.metadata as metadata
import logging
import os
import platform

from zrb.util.string.conversion import to_boolean


def _get_current_shell() -> str:
    if platform.system() == "Windows":
        return "PowerShell"
    current_shell = os.getenv("SHELL", "")
    if current_shell.endswith("zsh"):
        return "zsh"
    return "bash"


def _get_log_level(level: str) -> int:
    level = level.upper()
    log_levels = {
        "CRITICAL": logging.CRITICAL,  # 50
        "ERROR": logging.ERROR,  # 40
        "WARN": logging.WARNING,  # 30
        "WARNING": logging.WARNING,  # 30
        "INFO": logging.INFO,  # 20
        "DEBUG": logging.DEBUG,  # 10
        "NOTSET": logging.NOTSET,  # 0
    }
    if level in log_levels:
        return log_levels[level]
    return logging.WARNING


DEFAULT_SHELL = os.getenv("ZRB_SHELL", _get_current_shell())
DEFAULT_EDITOR = os.getenv("ZRB_EDITOR", "nano")
INIT_MODULES_STR = os.getenv("ZRB_INIT_MODULES", "")
INIT_MODULES = (
    [module.strip() for module in INIT_MODULES_STR.split(":") if module.strip() != ""]
    if INIT_MODULES_STR != ""
    else []
)
INIT_SCRIPTS_STR = os.getenv("ZRB_INIT_SCRIPTS", "")
INIT_SCRIPTS = (
    [script.strip() for script in INIT_SCRIPTS_STR.split(":") if script.strip() != ""]
    if INIT_SCRIPTS_STR != ""
    else []
)
LOGGING_LEVEL = _get_log_level(os.getenv("ZRB_LOGGING_LEVEL", "WARNING"))
LOAD_BUILTIN = to_boolean(os.getenv("ZRB_LOAD_BUILTIN", "1"))
WARN_UNRECOMMENDED_COMMAND = to_boolean(
    os.getenv("ZRB_WARN_UNRECOMMENDED_COMMAND", "1")
)
SESSION_LOG_DIR = os.getenv(
    "ZRB_SESSION_LOG_DIR", os.path.expanduser(os.path.join("~", ".zrb-session"))
)
TODO_DIR = os.getenv("ZRB_TODO_DIR", os.path.expanduser(os.path.join("~", "todo")))
TODO_VISUAL_FILTER = os.getenv("ZRB_TODO_FILTER", "")
TODO_RETENTION = os.getenv("ZRB_TODO_RETENTION", "2w")
VERSION = metadata.version("zrb")
WEB_HTTP_PORT = int(os.getenv("ZRB_WEB_HTTP_PORT", "21213"))
WEB_GUEST_USERNAME = os.getenv("ZRB_WEB_GUEST_USERNAME", "user")
WEB_SUPER_ADMIN_USERNAME = os.getenv("ZRB_WEB_SUPERADMIN_USERNAME", "admin")
WEB_SUPER_ADMIN_PASSWORD = os.getenv("ZRB_WEB_SUPERADMIN_PASSWORD", "admin")
WEB_ACCESS_TOKEN_COOKIE_NAME = os.getenv(
    "ZRB_WEB_ACCESS_TOKEN_COOKIE_NAME", "access_token"
)
WEB_REFRESH_TOKEN_COOKIE_NAME = os.getenv(
    "ZRB_WEB_REFRESH_TOKEN_COOKIE_NAME", "refresh_token"
)
WEB_SECRET_KEY = os.getenv("ZRB_WEB_SECRET", "zrb")
WEB_ENABLE_AUTH = to_boolean(os.getenv("ZRB_WEB_ENABLE_AUTH", "0"))
WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES = int(
    os.getenv("ZRB_WEB_ACCESS_TOKEN_EXPIRE_MINUTES", "30")
)
WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES = int(
    os.getenv("ZRB_WEB_REFRESH_TOKEN_EXPIRE_MINUTES", "60")
)

LLM_HISTORY_DIR = os.getenv(
    "ZRB_LLM_HISTORY_DIR", os.path.expanduser(os.path.join("~", ".zrb-llm-history"))
)
LLM_HISTORY_FILE = os.getenv(
    "ZRB_LLM_HISTORY_FILE", os.path.join(LLM_HISTORY_DIR, "history.json")
)
LLM_ALLOW_ACCESS_LOCAL_FILE = to_boolean(os.getenv("ZRB_LLM_ACCESS_LOCAL_FILE", "1"))
LLM_ALLOW_ACCESS_SHELL = to_boolean(os.getenv("ZRB_LLM_ACCESS_SHELL", "1"))
LLM_ALLOW_ACCESS_INTERNET = to_boolean(os.getenv("ZRB_LLM_ACCESS_INTERNET", "1"))
# RAG Configuration
RAG_EMBEDDING_API_KEY = os.getenv("ZRB_RAG_EMBEDDING_API_KEY", None)
RAG_EMBEDDING_BASE_URL = os.getenv("ZRB_RAG_EMBEDDING_BASE_URL", None)
RAG_EMBEDDING_MODEL = os.getenv("ZRB_RAG_EMBEDDING_MODEL", "text-embedding-ada-002")
RAG_CHUNK_SIZE = int(os.getenv("ZRB_RAG_CHUNK_SIZE", "1024"))
RAG_OVERLAP = int(os.getenv("ZRB_RAG_OVERLAP", "128"))
RAG_MAX_RESULT_COUNT = int(os.getenv("ZRB_RAG_MAX_RESULT_COUNT", "5"))
SERP_API_KEY = os.getenv("SERP_API_KEY", "")


BANNER = f"""
                bb
   zzzzz rr rr  bb
     zz  rrr  r bbbbbb
    zz   rr     bb   bb
   zzzzz rr     bbbbbb   {VERSION} Janggala
   _ _ . .  . _ .  _ . . .

Your Automation Powerhouse

☕ Donate at: https://stalchmst.com/donation
🐙 Submit issues/PR at: https://github.com/state-alchemists/zrb
🐤 Follow us at: https://twitter.com/zarubastalchmst
"""



from zrb.builtin.group import setup_latex_group
from zrb.builtin.setup.ubuntu import setup_ubuntu
from zrb.task.cmd_task import CmdTask

setup_latex_on_ubuntu = setup_latex_group.add_task(
    CmdTask(
        name="setup-latex-on-ubuntu",
        description="🐧 Setup LaTeX on Ubuntu",
        cmd=[
            "sudo apt install -y \\",
            "texlive-full texlive-latex-base texlive-fonts-recommended \\",
            "texlive-fonts-extra texlive-latex-extra",
        ],
        render_cmd=False,
    ),
    alias="ubuntu",
)
setup_ubuntu >> setup_latex_on_ubuntu



from my_app_name._zrb.config import APP_DIR
from my_app_name._zrb.group import app_group

from zrb.task.cmd_task import CmdTask

format_my_app_name_code = app_group.add_task(
    CmdTask(
        name="format-my-app-name-code",
        description="✨ Format Python code",
        cmd=[
            "isort . --profile black --force-grid-wrap 0",
            "black .",
        ],
        cwd=APP_DIR,
    ),
    alias="format",
)



from datetime import datetime, timedelta, timezone

from zrb.runner.web_config.config import WebConfig
from zrb.runner.web_schema.token import Token
from zrb.runner.web_util.user import get_user_by_credentials


def generate_tokens_by_credentials(
    web_config: WebConfig, username: str, password: str
) -> Token | None:
    if not web_config.enable_auth:
        user = web_config.default_user
    else:
        user = get_user_by_credentials(web_config, username, password)
    if user is None:
        return None
    access_token = _generate_access_token(web_config, user.username)
    refresh_token = _generate_refresh_token(web_config, user.username)
    return Token(
        access_token=access_token, refresh_token=refresh_token, token_type="bearer"
    )


def regenerate_tokens(web_config: WebConfig, refresh_token: str) -> Token:
    from fastapi import HTTPException
    from jose import jwt

    # Decode and validate token
    try:
        payload = jwt.decode(
            refresh_token,
            web_config.secret_key,
            options={"require_exp": True, "require_sub": True},
        )
    except Exception:
        raise HTTPException(status_code=401, detail="Invalid JWT token")
    if payload.get("type") != "refresh":
        raise HTTPException(status_code=401, detail="Invalid token type")
    username: str = payload.get("sub")
    if username is None:
        raise HTTPException(status_code=401, detail="Invalid refresh token")
    user = web_config.find_user_by_username(username)
    if user is None:
        raise HTTPException(status_code=401, detail="User not found")
    # Create new token
    new_access_token = _generate_access_token(web_config, username)
    new_refresh_token = _generate_refresh_token(web_config, username)
    return Token(
        access_token=new_access_token,
        refresh_token=new_refresh_token,
        token_type="bearer",
    )


def _generate_access_token(web_config: WebConfig, username: str) -> str:
    from jose import jwt

    expire = datetime.now(timezone.utc) + timedelta(
        minutes=web_config.access_token_expire_minutes
    )
    to_encode = {"sub": username, "exp": expire, "type": "access"}
    return jwt.encode(to_encode, web_config.secret_key)


def _generate_refresh_token(web_config: WebConfig, username: str) -> str:
    from jose import jwt

    expire = datetime.now(timezone.utc) + timedelta(
        minutes=web_config.refresh_token_expire_minutes
    )
    to_encode = {"sub": username, "exp": expire, "type": "refresh"}
    return jwt.encode(to_encode, web_config.secret_key)



import os

from my_app_name._zrb.config import ACTIVATE_VENV_SCRIPT, APP_DIR

from zrb import CmdTask

create_venv = CmdTask(
    name="create-my-app-name-venv",
    cwd=APP_DIR,
    cmd="python -m venv .venv",
    execute_condition=lambda _: not os.path.isdir(os.path.join(APP_DIR, ".venv")),
)

prepare_venv = CmdTask(
    name="prepare-my-app-name-venv",
    cmd=[ACTIVATE_VENV_SCRIPT, "pip install -r requirements.txt"],
    cwd=APP_DIR,
)
create_venv >> prepare_venv



def prepend_code_to_module(original_code: str, new_code: str) -> str:
    lines = original_code.splitlines()
    last_import_index = -1
    for i, line in enumerate(lines):
        stripped_line = line.strip()
        if stripped_line.startswith("import") or stripped_line.startswith("from"):
            last_import_index = i
        else:
            break
    lines.insert(last_import_index + 1, new_code)
    return "\n".join(lines)



import os

from pydantic_ai.models import Model
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers import Provider
from pydantic_ai.providers.openai import OpenAIProvider

DEFAULT_SYSTEM_PROMPT = """
You have access to tools.
Your goal is to provide insightful and accurate information based on user queries.
Follow these instructions precisely:
1. ALWAYS use available tools to gather information BEFORE asking the user questions.
2. For tools that require arguments: provide arguments in valid JSON format.
3. For tools with no args: call the tool without args. Do NOT pass "" or {}.
4. NEVER pass arguments to tools that don't accept parameters.
5. NEVER ask users for information obtainable through tools.
6. Use tools in a logical sequence until you have sufficient information.
7. If a tool call fails, check if you're passing arguments in the correct format.
   Consider alternative strategies if the issue persists.
8. Only after exhausting relevant tools should you request clarification.
9. Understand the context of user queries to provide relevant and accurate responses.
10. Engage with users in a conversational manner once the necessary information is gathered.
11. Adapt to different query types or scenarios to improve flexibility and effectiveness.
""".strip()

DEFAULT_PERSONA = """
You are an expert in various fields including technology, science, history, and more.
""".strip()


class LLMConfig:

    def __init__(
        self,
        default_model_name: str | None = None,
        default_base_url: str | None = None,
        default_api_key: str | None = None,
        default_persona: str | None = None,
        default_system_prompt: str | None = None,
    ):
        self._default_model_name = (
            default_model_name
            if default_model_name is not None
            else os.getenv("ZRB_LLM_MODEL", None)
        )
        self._default_model_base_url = (
            default_base_url
            if default_base_url is not None
            else os.getenv("ZRB_LLM_BASE_URL", None)
        )
        self._default_model_api_key = (
            default_api_key
            if default_api_key is not None
            else os.getenv("ZRB_LLM_API_KEY", None)
        )
        self._default_system_prompt = (
            default_system_prompt
            if default_system_prompt is not None
            else os.getenv("ZRB_LLM_SYSTEM_PROMPT", None)
        )
        self._default_persona = (
            default_persona
            if default_persona is not None
            else os.getenv("ZRB_LLM_PERSONA", None)
        )
        self._default_provider = None
        self._default_model = None

    def _get_model_name(self) -> str | None:
        return (
            self._default_model_name if self._default_model_name is not None else None
        )

    def get_default_model_provider(self) -> Provider | str:
        if self._default_provider is not None:
            return self._default_provider
        if self._default_model_base_url is None and self._default_model_api_key is None:
            return "openai"
        return OpenAIProvider(
            base_url=self._default_model_base_url, api_key=self._default_model_api_key
        )

    def get_default_system_prompt(self) -> str:
        system_prompt = (
            DEFAULT_SYSTEM_PROMPT
            if self._default_system_prompt is None
            else self._default_system_prompt
        )
        persona = (
            DEFAULT_PERSONA if self._default_persona is None else self._default_persona
        )
        if persona is not None:
            return f"{persona}\n{system_prompt}"
        return system_prompt

    def get_default_model(self) -> Model | str | None:
        if self._default_model is not None:
            return self._default_model
        model_name = self._get_model_name()
        if model_name is None:
            return None
        return OpenAIModel(
            model_name=model_name,
            provider=self.get_default_model_provider(),
        )

    def set_default_persona(self, persona: str):
        self._default_persona = persona

    def set_default_system_prompt(self, system_prompt: str):
        self._default_system_prompt = system_prompt

    def set_default_model_name(self, model_name: str):
        self._default_model_name = model_name

    def set_default_model_api_key(self, model_api_key: str):
        self._default_model_api_key = model_api_key

    def set_default_model_base_url(self, model_base_url: str):
        self._default_model_base_url = model_base_url

    def set_default_provider(self, provider: Provider | str):
        self._default_provider = provider

    def set_default_model(self, model: Model | str | None):
        self._default_model = model


llm_config = LLMConfig()



import os
from collections.abc import Callable
from typing import Any

from pydantic import BaseModel

from zrb.util.cmd.command import run_command


class DiffResult(BaseModel):
    created: list[str]
    removed: list[str]
    updated: list[str]


async def get_diff(
    repo_dir: str,
    source_commit: str,
    current_commit: str,
    print_method: Callable[..., Any] = print,
) -> DiffResult:
    cmd_result, exit_code = await run_command(
        cmd=["git", "diff", source_commit, current_commit],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    lines = cmd_result.output.strip().split("\n")
    diff: dict[str, dict[str, bool]] = {}
    for line in lines:
        if not line.startswith("---") and not line.startswith("+++"):
            continue
        if line[4:6] != "a/" and line[4:6] != "b/":
            continue
        # line should contains something like `--- a/some-file.txt`
        file = line[6:]
        if file not in diff:
            diff[file] = {"plus": False, "minus": False}
        if line.startswith("---"):
            diff[file]["minus"] = True
        if line.startswith("+++"):
            diff[file]["plus"] = True
    return DiffResult(
        created=[
            file for file, state in diff.items() if state["plus"] and not state["minus"]
        ],
        removed=[
            file for file, state in diff.items() if not state["plus"] and state["minus"]
        ],
        updated=[
            file for file, state in diff.items() if state["plus"] and state["minus"]
        ],
    )


async def get_repo_dir(print_method: Callable[..., Any] = print) -> str:
    cmd_result, exit_code = await run_command(
        cmd=["git", "rev-parse", "--show-toplevel"],
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    return os.path.abspath(cmd_result.output.strip())


async def get_current_branch(
    repo_dir: str, print_method: Callable[..., Any] = print
) -> str:
    cmd_result, exit_code = await run_command(
        cmd=["git", "rev-parse", "--abbrev-ref", "HEAD"],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    return cmd_result.output.strip()


async def get_branches(
    repo_dir: str, print_method: Callable[..., Any] = print
) -> list[str]:
    cmd_result, exit_code = await run_command(
        cmd=["git", "rev-parse", "--abbrev-ref", "HEAD"],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    return [
        branch.lstrip("*").strip() for branch in cmd_result.output.strip().split("\n")
    ]


async def delete_branch(
    repo_dir: str, branch_name: str, print_method: Callable[..., Any] = print
) -> str:
    cmd_result, exit_code = await run_command(
        cmd=["git", "branch", "-D", branch_name],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    return cmd_result.output.strip()


async def add(repo_dir: str, print_method: Callable[..., Any] = print):
    _, exit_code = await run_command(
        cmd=["git", "add", ".", "-A"],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")


async def commit(
    repo_dir: str, message: str, print_method: Callable[..., Any] = print
) -> str:
    cmd_result, exit_code = await run_command(
        cmd=["git", "commit", "-m", message],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        ignored_error_message = "nothing to commit, working tree clean"
        if (
            ignored_error_message not in cmd_result.error
            and ignored_error_message not in cmd_result.output
        ):
            raise Exception(f"Non zero exit code: {exit_code}")


async def pull(
    repo_dir: str, remote: str, branch: str, print_method: Callable[..., Any] = print
) -> str:
    _, exit_code = await run_command(
        cmd=["git", "pull", remote, branch],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")


async def push(
    repo_dir: str, remote: str, branch: str, print_method: Callable[..., Any] = print
) -> str:
    _, exit_code = await run_command(
        cmd=["git", "push", "-u", remote, branch],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")



import json
from collections.abc import Callable
from typing import Annotated


async def open_web_page(url: str) -> str:
    """Get content from a web page using a headless browser."""

    async def get_page_content(page_url: str):
        user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"  # noqa
        try:
            from playwright.async_api import async_playwright

            async with async_playwright() as p:
                browser = await p.chromium.launch(headless=True)
                page = await browser.new_page()
                await page.set_extra_http_headers({"User-Agent": user_agent})
                try:
                    # Navigate to the URL with a timeout of 30 seconds
                    await page.goto(page_url, wait_until="networkidle", timeout=30000)
                    # Wait for the content to load
                    await page.wait_for_load_state("domcontentloaded")
                    # Get the page content
                    content = await page.content()
                    # Extract all links from the page
                    links = await page.eval_on_selector_all(
                        "a[href]",
                        """
                        (elements) => elements.map(el => {
                            const href = el.getAttribute('href');
                            if (href && !href.startsWith('#') && !href.startsWith('/')) {
                                return href;
                            }
                            return null;
                        }).filter(href => href !== null)
                    """,
                    )
                    return {"content": content, "links_on_page": links}
                finally:
                    await browser.close()
        except ImportError:
            import requests

            response = requests.get(url, headers={"User-Agent": user_agent})
            if response.status_code != 200:
                msg = f"Unable to retrieve search results. Status code: {response.status_code}"
                raise Exception(msg)
            return {"content": response.text, "links_on_page": []}

    result = await get_page_content(url)
    # Parse the HTML content
    return json.dumps(parse_html_text(result["content"]))


def create_search_internet_tool(serp_api_key: str) -> Callable[[str, int], str]:
    def search_internet(
        query: Annotated[str, "Search query"],
        num_results: Annotated[int, "Search result count, by default 10"] = 10,
    ) -> str:
        """Search factual information from the internet by using Google."""
        import requests

        response = requests.get(
            "https://serpapi.com/search",
            headers={
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"  # noqa
            },
            params={
                "q": query,
                "num": num_results,
                "hl": "en",
                "safe": "off",
                "api_key": serp_api_key,
            },
        )
        if response.status_code != 200:
            raise Exception(
                f"Error: Unable to retrieve search results (status code: {response.status_code})"  # noqa
            )
        return json.dumps(parse_html_text(response.text))

    return search_internet


def search_wikipedia(query: Annotated[str, "Search query"]) -> str:
    """Search on wikipedia"""
    import requests

    params = {"action": "query", "list": "search", "srsearch": query, "format": "json"}
    response = requests.get("https://en.wikipedia.org/w/api.php", params=params)
    return response.json()


def search_arxiv(
    query: Annotated[str, "Search query"],
    num_results: Annotated[int, "Search result count, by default 10"] = 10,
) -> str:
    """Search on Arxiv"""
    import requests

    params = {"search_query": f"all:{query}", "start": 0, "max_results": num_results}
    response = requests.get("http://export.arxiv.org/api/query", params=params)
    return response.content


def parse_html_text(html_text: str) -> dict[str, str]:
    from bs4 import BeautifulSoup

    ignored_tags = [
        "script",
        "link",
        "meta",
        "style",
        "code",
        "footer",
        "nav",
        "header",
        "aside",
    ]
    soup = BeautifulSoup(html_text, "html.parser")
    links = []
    for anchor in soup.find_all("a"):
        if not anchor or "href" not in anchor.attrs:
            continue
        link: str = anchor["href"]
        if link.startswith("#") or link.startswith("/"):
            continue
        links.append(link)
    for tag in soup(ignored_tags):
        tag.decompose()
    html_text = soup.get_text(separator=" ", strip=True)
    return {"content": html_text, "links_on_page": links}



from typing import Callable

from zrb.runner.web_schema.user import User
from zrb.task.any_task import AnyTask


class WebConfig:
    def __init__(
        self,
        port: int,
        secret_key: str,
        access_token_expire_minutes: int,
        refresh_token_expire_minutes: int,
        access_token_cookie_name: str,
        refresh_token_cookie_name: str,
        enable_auth: bool,
        super_admin_username: str,
        super_admin_password: str,
        guest_username: str,
        guest_accessible_tasks: list[AnyTask | str] = [],
        find_user_by_username: Callable[[str], User | None] | None = None,
    ):
        self.secret_key = secret_key
        self.access_token_expire_minutes = access_token_expire_minutes
        self.refresh_token_expire_minutes = refresh_token_expire_minutes
        self.access_token_cookie_name = access_token_cookie_name
        self.refresh_token_cookie_name = refresh_token_cookie_name
        self.enable_auth = enable_auth
        self.port = port
        self._user_list = []
        self.super_admin_username = super_admin_username
        self.super_admin_password = super_admin_password
        self.guest_username = guest_username
        self.guest_accessible_tasks = guest_accessible_tasks
        self._find_user_by_username = find_user_by_username

    @property
    def default_user(self) -> User:
        if self.enable_auth:
            return User(
                username=self.guest_username,
                password="",
                is_guest=True,
                accessible_tasks=self.guest_accessible_tasks,
            )
        return User(
            username=self.guest_username,
            password="",
            is_guest=True,
            is_super_admin=True,
        )

    @property
    def super_admin(self) -> User:
        return User(
            username=self.super_admin_username,
            password=self.super_admin_password,
            is_super_admin=True,
        )

    @property
    def user_list(self) -> list[User]:
        if not self.enable_auth:
            return [self.default_user]
        return self._user_list + [self.super_admin, self.default_user]

    def set_guest_accessible_tasks(self, tasks: list[AnyTask | str]):
        self.guest_accessible_tasks = tasks

    def set_find_user_by_username(
        self, find_user_by_username: Callable[[str], User | None]
    ):
        self._find_user_by_username = find_user_by_username

    def append_user(self, user: User):
        duplicates = [
            existing_user
            for existing_user in self.user_list
            if existing_user.username == user.username
        ]
        if len(duplicates) > 0:
            raise ValueError(f"User already exists {user.username}")
        self._user_list.append(user)

    def find_user_by_username(self, username: str) -> User | None:
        user = None
        if self._find_user_by_username is not None:
            user = self._find_user_by_username(username)
        if user is None:
            user = next((u for u in self.user_list if u.username == username), None)
        return user



import datetime
import logging
import sys
from typing import Any, TextIO

from zrb.context.any_context import AnyContext
from zrb.context.any_shared_context import AnySharedContext
from zrb.dot_dict.dot_dict import DotDict
from zrb.session.any_session import AnySession
from zrb.util.cli.style import (
    remove_style,
    stylize,
    stylize_error,
    stylize_log,
    stylize_warning,
)
from zrb.util.string.conversion import to_boolean


class Context(AnyContext):
    def __init__(
        self, shared_ctx: AnySharedContext, task_name: str, color: int, icon: str
    ):
        self._shared_ctx = shared_ctx
        self._env = shared_ctx.env.copy()
        self._task_name = task_name
        self._color = color
        self._icon = icon
        self._attempt = 0
        self._max_attempt = 0

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"<{class_name} shared_ctx={self._shared_ctx}>"

    @property
    def input(self) -> DotDict:
        return self._shared_ctx.input

    @property
    def env(self) -> DotDict:
        return self._env

    @property
    def args(self) -> list[Any]:
        return self._shared_ctx.args

    @property
    def xcom(self) -> DotDict:
        return self._shared_ctx.xcom

    @property
    def shared_log(self) -> list[str]:
        return self._shared_ctx.shared_log

    @property
    def session(self) -> AnySession | None:
        return self._shared_ctx._session

    def update_task_env(self, task_env: dict[str, str]):
        self._env.update(task_env)

    def append_to_shared_log(self, message: str):
        self._shared_ctx.append_to_shared_log(message)

    def set_session(self, session: AnySession):
        self._shared_ctx.set_session(session)

    def set_attempt(self, attempt: int):
        self._attempt = attempt

    def set_max_attempt(self, max_attempt: int):
        self._max_attempt = max_attempt

    def get_logging_level(self) -> int:
        return self._shared_ctx.get_logging_level()

    def render(self, template: str) -> str:
        return self._shared_ctx.render(template=template)

    def render_bool(self, template: str | bool) -> bool:
        if isinstance(template, bool):
            return template
        return to_boolean(self.render(template))

    def render_int(self, template: str | int) -> int:
        if isinstance(template, int):
            return template
        return int(self.render(template))

    def render_float(self, template: str | float) -> float:
        if isinstance(template, float):
            return template
        return float(self.render(template))

    def print(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
        plain: bool = False,
    ):
        sep = " " if sep is None else sep
        message = sep.join([f"{value}" for value in values])
        if plain:
            # self.append_to_shared_log(remove_style(message))
            print(message, sep=sep, end=end, file=file, flush=flush)
            return
        color = self._color
        icon = self._icon
        max_name_length = max(len(name) + len(icon) for name in self.session.task_names)
        styled_task_name = f"{icon} {self._task_name}"
        padded_styled_task_name = styled_task_name.rjust(max_name_length + 1)
        if self._attempt == 0:
            attempt_status = "".ljust(5)
        else:
            attempt_status = f"{self._attempt}/{self._max_attempt}".ljust(5)
        now = datetime.datetime.now()
        formatted_time = now.strftime("%y%m%d %H:%M:%S.%f")[:19] + " "
        prefix = f"{formatted_time}{attempt_status} {padded_styled_task_name} ⬤ "
        self.append_to_shared_log(remove_style(f"{prefix} {message}"))
        stylized_prefix = stylize(prefix, color=color)
        print(f"{stylized_prefix} {message}", sep=sep, end=end, file=file, flush=flush)

    def log_debug(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        if self._shared_ctx.get_logging_level() <= logging.DEBUG:
            sep = " " if sep is None else sep
            message = sep.join([f"{value}" for value in values])
            stylized_message = stylize_log(f"[DEBUG] {message}")
            self.print(stylized_message, sep=sep, end=end, file=file, flush=flush)

    def log_info(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        if self._shared_ctx.get_logging_level() <= logging.INFO:
            sep = " " if sep is None else sep
            message = sep.join([f"{value}" for value in values])
            stylized_message = stylize_log(f"[INFO] {message}")
            self.print(stylized_message, sep=sep, end=end, file=file, flush=flush)

    def log_warning(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        if self._shared_ctx.get_logging_level() <= logging.INFO:
            sep = " " if sep is None else sep
            message = sep.join([f"{value}" for value in values])
            stylized_message = stylize_warning(f"[WARNING] {message}")
            self.print(stylized_message, sep=sep, end=end, file=file, flush=flush)

    def log_error(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        if self._shared_ctx.get_logging_level() <= logging.ERROR:
            sep = " " if sep is None else sep
            message = sep.join([f"{value}" for value in values])
            stylized_message = stylize_error(f"[ERROR] {message}")
            self.print(stylized_message, sep=sep, end=end, file=file, flush=flush)

    def log_critical(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        if self._shared_ctx.get_logging_level() <= logging.CRITICAL:
            sep = " " if sep is None else sep
            message = sep.join([f"{value}" for value in values])
            stylized_message = stylize_error(f"[CRITICAL] {message}")
            self.print(stylized_message, sep=sep, end=end, file=file, flush=flush)



import os
import re


def read_file(file_path: str, replace_map: dict[str, str] = {}) -> str:
    with open(file_path, "r", encoding="utf-8") as f:
        content = f.read()
    for key, val in replace_map.items():
        content = content.replace(key, val)
    return content


def write_file(file_path: str, content: str | list[str]):
    if isinstance(content, list):
        content = "\n".join([line for line in content if line is not None])
    dir_path = os.path.dirname(file_path)
    os.makedirs(dir_path, exist_ok=True)
    should_add_eol = content.endswith("\n")
    # Remove trailing newlines, but keep one if the file originally ended up with newline
    content = re.sub(r"\n{3,}$", "\n\n", content)
    content = content.rstrip("\n")
    if should_add_eol:
        content += "\n"
    with open(file_path, "w") as f:
        f.write(content)



from collections.abc import Callable

from zrb.attr.type import IntAttr, StrAttr
from zrb.context.any_context import AnyContext
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.task.any_task import AnyTask
from zrb.task.cmd_task import CmdTask
from zrb.util.attr import get_str_attr


class RsyncTask(CmdTask):
    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput] | AnyInput | None = None,
        env: list[AnyEnv] | AnyEnv | None = None,
        shell: StrAttr | None = None,
        auto_render_shell: bool = True,
        remote_host: StrAttr | None = None,
        render_remote_host: bool = True,
        remote_port: IntAttr | None = None,
        render_remote_port: bool = True,
        remote_user: StrAttr | None = None,
        render_remote_user: bool = True,
        remote_password: StrAttr | None = None,
        render_remote_password: bool = True,
        remote_ssh_key: StrAttr | None = None,
        render_remote_ssh_key: bool = True,
        remote_source_path: StrAttr | None = None,
        render_remote_source_path: bool = True,
        remote_destination_path: StrAttr | None = None,
        render_remote_destination_path: bool = True,
        local_source_path: StrAttr | None = None,
        render_local_source_path: bool = True,
        local_destination_path: StrAttr | None = None,
        render_local_destination_path: bool = True,
        cwd: str | None = None,
        render_cwd: bool = True,
        plain_print: bool = False,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        execute_condition: bool | str | Callable[[AnyContext], bool] = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
        successor: list[AnyTask] | AnyTask | None = None,
    ):
        super().__init__(
            name=name,
            color=color,
            icon=icon,
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            shell=shell,
            render_shell=auto_render_shell,
            remote_host=remote_host,
            render_remote_host=render_remote_host,
            remote_port=remote_port,
            auto_render_remote_port=render_remote_port,
            remote_user=remote_user,
            render_remote_user=render_remote_user,
            remote_password=remote_password,
            render_remote_password=render_remote_password,
            remote_ssh_key=remote_ssh_key,
            render_remote_ssh_key=render_remote_ssh_key,
            cwd=cwd,
            render_cwd=render_cwd,
            plain_print=plain_print,
            max_output_line=max_output_line,
            max_error_line=max_error_line,
            execute_condition=execute_condition,
            retries=retries,
            retry_period=retry_period,
            readiness_check=readiness_check,
            upstream=upstream,
            fallback=fallback,
            successor=successor,
        )
        self._remote_source_path = remote_source_path
        self._render_remote_source_path = render_remote_source_path
        self._remote_destination_path = remote_destination_path
        self._render_remote_destination_path = render_remote_destination_path
        self._local_source_path = local_source_path
        self._render_local_source_path = render_local_source_path
        self._local_destination_path = local_destination_path
        self._render_local_destination_path = render_local_destination_path

    def _get_source_path(self, ctx: AnyContext) -> str:
        local_source_path = self._get_local_source_path(ctx)
        if local_source_path != "":
            return local_source_path
        remote_source_path = self._get_remote_source_path(ctx)
        host = self._get_remote_host(ctx)
        user = self._get_remote_user(ctx)
        return f"{user}@{host}:{remote_source_path}"

    def _get_destination_path(self, ctx: AnyContext) -> str:
        local_destination_path = self._get_local_destination_path(ctx)
        if local_destination_path != "":
            return local_destination_path
        remote_destination_path = self._get_remote_destination_path(ctx)
        host = self._get_remote_host(ctx)
        user = self._get_remote_user(ctx)
        return f"{user}@{host}:{remote_destination_path}"

    def _get_remote_source_path(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx,
            self._remote_source_path,
            "",
            auto_render=self._render_remote_source_path,
        )

    def _get_remote_destination_path(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx,
            self._remote_destination_path,
            "",
            auto_render=self._render_remote_destination_path,
        )

    def _get_local_source_path(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx,
            self._local_source_path,
            "",
            auto_render=self._render_local_source_path,
        )

    def _get_local_destination_path(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx,
            self._local_destination_path,
            "",
            auto_render=self._render_local_destination_path,
        )

    def _get_cmd_script(self, ctx: AnyContext) -> str:
        port = self._get_remote_port(ctx)
        password = self._get_remote_password(ctx)
        key = self._get_remote_ssh_key(ctx)
        src = self._get_source_path(ctx)
        dst = self._get_destination_path(ctx)
        if key != "" and password != "":
            return f'sshpass -p "$_ZRB_SSH_PASSWORD" rsync --mkpath -avz -e "ssh -i {key} -p {port}" {src} {dst}'  # noqa
        if key != "":
            return f'rsync --mkpath -avz -e "ssh -i {key} -p {port}" {src} {dst}'
        if password != "":
            return f'sshpass -p "$_ZRB_SSH_PASSWORD" rsync --mkpath -avz -e "ssh -p {port}" {src} {dst}'  # noqa
        return f'rsync --mkpath -avz -e "ssh -p {port}" {src} {dst}'



import asyncio
import os
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any

from zrb.context.shared_context import SharedContext
from zrb.group.any_group import AnyGroup
from zrb.runner.web_config.config import WebConfig
from zrb.runner.web_schema.session import NewSessionResponse
from zrb.runner.web_util.user import get_user_from_request
from zrb.session.session import Session
from zrb.session_state_log.session_state_log import SessionStateLog, SessionStateLogList
from zrb.session_state_logger.any_session_state_logger import AnySessionStateLogger
from zrb.task.any_task import AnyTask
from zrb.util.group import NodeNotFoundError, extract_node_from_args, get_node_path

if TYPE_CHECKING:
    # We want fastapi to only be loaded when necessary to decrease footprint

    from fastapi import FastAPI


def serve_task_session_api(
    app: "FastAPI",
    root_group: AnyGroup,
    web_config: WebConfig,
    session_state_logger: AnySessionStateLogger,
    coroutines: list,
) -> None:
    from fastapi import Query, Request
    from fastapi.responses import JSONResponse

    @app.post("/api/v1/task-sessions/{path:path}")
    async def create_new_task_session_api(
        path: str,
        request: Request,
        inputs: dict[str, Any],
    ) -> NewSessionResponse:
        """
        Creating new session
        """
        user = await get_user_from_request(web_config, request)
        args = path.strip("/").split("/")
        try:
            task, _, residual_args = extract_node_from_args(root_group, args)
        except NodeNotFoundError:
            return JSONResponse(content={"detail": "Not found"}, status_code=404)
        if isinstance(task, AnyTask):
            if not user.can_access_task(task):
                return JSONResponse(content={"detail": "Forbidden"}, status_code=403)
            session_name = residual_args[0] if residual_args else None
            if not session_name:
                shared_ctx = SharedContext(env=dict(os.environ))
                session = Session(shared_ctx=shared_ctx, root_group=root_group)
                coro = asyncio.create_task(task.async_run(session, str_kwargs=inputs))
                coroutines.append(coro)
                coro.add_done_callback(lambda coro: coroutines.remove(coro))
                return NewSessionResponse(session_name=session.name)
        return JSONResponse(content={"detail": "Not found"}, status_code=404)

    @app.get(
        "/api/v1/task-sessions/{path:path}",
        response_model=SessionStateLog | SessionStateLogList,
    )
    async def get_task_session_api(
        path: str,
        request: Request,
        min_start_query: str = Query(default=None, alias="from"),
        max_start_query: str = Query(default=None, alias="to"),
        page: int = Query(default=0, alias="page"),
        limit: int = Query(default=10, alias="limit"),
    ) -> SessionStateLog | SessionStateLogList:
        """
        Getting existing session or sessions
        """
        user = await get_user_from_request(web_config, request)
        args = path.strip("/").split("/")
        try:
            task, _, residual_args = extract_node_from_args(root_group, args)
        except NodeNotFoundError:
            return JSONResponse(content={"detail": "Not found"}, status_code=404)
        if isinstance(task, AnyTask) and residual_args:
            if not user.can_access_task(task):
                return JSONResponse(content={"detail": "Forbidden"}, status_code=403)
            if residual_args[0] == "list":
                task_path = get_node_path(root_group, task)
                max_start_time = (
                    datetime.now()
                    if max_start_query is None
                    else datetime.strptime(max_start_query, "%Y-%m-%d %H:%M:%S")
                )
                min_start_time = (
                    max_start_time - timedelta(hours=1)
                    if min_start_query is None
                    else datetime.strptime(min_start_query, "%Y-%m-%d %H:%M:%S")
                )
                return sanitize_session_state_log_list(
                    task,
                    session_state_logger.list(
                        task_path, min_start_time, max_start_time, page, limit
                    ),
                )
            else:
                return sanitize_session_state_log(
                    task, session_state_logger.read(residual_args[0])
                )
        return JSONResponse(content={"detail": "Not found"}, status_code=404)


def sanitize_session_state_log_list(
    task: AnyTask, session_state_log_list: SessionStateLogList
) -> SessionStateLogList:
    return SessionStateLogList(
        total=session_state_log_list.total,
        data=[
            sanitize_session_state_log(task, data)
            for data in session_state_log_list.data
        ],
    )


def sanitize_session_state_log(
    task: AnyTask, session_state_log: SessionStateLog
) -> SessionStateLog:
    """
    In session, we create snake_case aliases of inputs.
    The purpose was to increase ergonomics, so that user can use `input.system_prompt`
    instead of `input["system-prompt"]`
    However, when we serve the session through HTTP API,
    we only want to show the original input names.
    """
    enhanced_inputs = session_state_log.input
    real_inputs = {}
    for real_input in task.inputs:
        real_input_name = real_input.name
        real_inputs[real_input_name] = enhanced_inputs[real_input_name]
    return SessionStateLog(
        name=session_state_log.name,
        start_time=session_state_log.start_time,
        main_task_name=session_state_log.main_task_name,
        path=session_state_log.path,
        input=real_inputs,
        final_result=session_state_log.final_result,
        finished=session_state_log.finished,
        log=session_state_log.log,
        task_status=session_state_log.task_status,
    )



import os

from zrb.builtin.group import setup_group
from zrb.builtin.setup.common_input import package_manager_input, use_sudo_input
from zrb.builtin.setup.zsh.zsh_helper import get_install_zsh_cmd
from zrb.context.any_context import AnyContext
from zrb.input.str_input import StrInput
from zrb.task.cmd_task import CmdTask
from zrb.task.make_task import make_task
from zrb.util.file import read_file, write_file

install_zsh = CmdTask(
    name="install-zsh",
    input=[package_manager_input, use_sudo_input],
    cmd=get_install_zsh_cmd,
)

install_omz = CmdTask(
    name="install-omz",
    cmd='sh -c "$(curl -fsSL https://raw.githubusercontent.com/ohmyzsh/ohmyzsh/master/tools/install.sh)"',  # noqa
    upstream=install_zsh,
)

install_zinit = CmdTask(
    name="install-zinit",
    cmd='bash -c "$(curl --fail --show-error --silent --location https://raw.githubusercontent.com/zdharma-continuum/zinit/HEAD/scripts/install.sh)"',  # noqa
    upstream=install_zsh,
)


@make_task(
    name="setup-zsh",
    input=StrInput(
        name="zsh-config",
        description="zsh config file",
        prompt="zsh config file",
        default="~/.zshrc",
    ),
    upstream=[install_omz, install_zinit],
    description="💻 Setup `zsh`.",
    group=setup_group,
    alias="zsh",
)
def setup_zsh(ctx: AnyContext):
    zsh_config = read_file(os.path.join(os.path.dirname(__file__), "zsh_config.sh"))
    zsh_config_file = os.path.expanduser(ctx.input["zsh-config"])
    # Make sure config file exists
    if not os.path.isfile(zsh_config_file):
        write_file(zsh_config_file, "")
    content = read_file(zsh_config_file)
    if zsh_config in content:
        return
    # Write config
    write_file(zsh_config_file, [content, zsh_config, ""])
    ctx.print("Setup complete, restart your terminal to continue")



import fnmatch
import json
import os
import re
from typing import Any, Dict, List, Optional

from zrb.util.file import read_file as _read_file
from zrb.util.file import write_file as _write_file


def list_files(
    path: str = ".", recursive: bool = True, include_hidden: bool = False
) -> str:
    """
    Request to list files and directories within the specified directory.
    If recursive is true, it will list all files and directories recursively.
    If recursive is false or not provided, it will only list the top-level contents.
    Args:
        path: (required) The path of the directory to list contents for (relative to the CWD)
        recursive: (optional) Whether to list files recursively.
            Use true for recursive listing, false or omit for top-level only.
        include_hidden: (optional) Whether to include hidden files/directories.
            Defaults to False (exclude hidden files).
    Returns:
        A JSON string containing a list of file paths or an error message.
        Example success: '{"files": ["file1.txt", "subdir/file2.py"]}'
        Example error: '{"error": "Error listing files: [Errno 2] No such file..."}'
    """
    all_files: List[str] = []
    abs_path = os.path.abspath(path)
    try:
        if recursive:
            for root, dirs, files in os.walk(abs_path):
                # Skip hidden directories (like .git) for performance and relevance
                dirs[:] = [d for d in dirs if include_hidden or not _is_hidden(d)]
                for filename in files:
                    # Skip hidden files
                    if include_hidden or not _is_hidden(filename):
                        all_files.append(os.path.join(root, filename))
        else:
            # Non-recursive listing (top-level only)
            for item in os.listdir(abs_path):
                full_path = os.path.join(abs_path, item)
                # Include both files and directories if not recursive
                if include_hidden or not _is_hidden(
                    item
                ):  # Skip hidden items unless included
                    all_files.append(full_path)

        # Return paths relative to the original path requested
        try:
            rel_files = [
                os.path.relpath(f, os.path.dirname(abs_path)) for f in all_files
            ]
            return json.dumps({"files": sorted(rel_files)})
        except (
            ValueError
        ) as e:  # Handle case where path is '.' and abs_path is CWD root
            if "path is on mount '" in str(e) and "' which is not on mount '" in str(e):
                # If paths are on different mounts, just use absolute paths
                rel_files = all_files
                return json.dumps({"files": sorted(rel_files)})
            raise
    except Exception as e:
        raise Exception(f"Error listing files in {path}: {e}")


def _is_hidden(path: str) -> bool:
    """Check if path is hidden (starts with '.')."""
    return os.path.basename(path).startswith(".")


def read_from_file(
    path: str,
    start_line: Optional[int] = None,
    end_line: Optional[int] = None,
) -> str:
    """
    Request to read the contents of a file at the specified path. Use this when you need
    to examine the contents of an existing file you do not know the contents of, for example
    to analyze code, review text files, or extract information from configuration files.
    The output includes line numbers prefixed to each line (e.g. "1 | const x = 1"),
    making it easier to reference specific lines when creating diffs or discussing code.
    By specifying start_line and end_line parameters, you can efficiently read specific
    portions of large files without loading the entire file into memory. Automatically
    extracts raw text from PDF and DOCX files. May not be suitable for other types of
    binary files, as it returns the raw content as a string.
    Args:
        path: (required) The path of the file to read (relative to the CWD)
        start_line: (optional) The starting line number to read from (1-based).
            If not provided, it starts from the beginning of the file.
        end_line: (optional) The ending line number to read to (1-based, inclusive).
            If not provided, it reads to the end of the file.
    Returns:
        A JSON string containing the file path, content, and line range, or an error.
        Example success: '{"path": "f.py", "content": "...", "start_line": 1, "end_line": 2}'
        Example error: '{"error": "File not found: data.txt"}'
    """
    try:
        abs_path = os.path.abspath(path)
        # Check if file exists
        if not os.path.exists(abs_path):
            return json.dumps({"error": f"File {path} does not exist"})
        content = _read_file(abs_path)
        lines = content.splitlines()
        total_lines = len(lines)
        # Adjust line indices (convert from 1-based to 0-based)
        start_idx = (start_line - 1) if start_line is not None else 0
        end_idx = end_line if end_line is not None else total_lines
        # Validate indices
        if start_idx < 0:
            start_idx = 0
        if end_idx > total_lines:
            end_idx = total_lines
        if start_idx > end_idx:
            start_idx = end_idx
        # Select the lines for the result
        selected_lines = lines[start_idx:end_idx]
        content_result = "\n".join(selected_lines)
        return json.dumps(
            {
                "path": path,
                "content": content_result,
                "start_line": start_idx + 1,  # Convert back to 1-based for output
                "end_line": end_idx,  # end_idx is already exclusive upper bound
                "total_lines": total_lines,
            }
        )
    except Exception as e:
        raise Exception(f"Error reading file {path}: {e}")


def write_to_file(path: str, content: str, line_count: int) -> str:
    """
    Request to write full content to a file at the specified path. If the file exists,
    it will be overwritten with the provided content. If the file doesn't exist,
    it will be created. This tool will automatically create any directories needed
    to write the file.
    Args:
        path: (required) The path of the file to write to (relative to the CWD)
        content: (required) The content to write to the file. ALWAYS provide the COMPLETE
            intended content of the file, without any truncation or omissions. You MUST
            include ALL parts of the file, even if they haven't been modified. Do NOT
            include the line numbers in the content though, just the actual content
            of the file.
        line_count: (required) The number of lines in the file. Make sure to compute
            this based on the actual content of the file, not the number of lines
            in the content you're providing.
    Returns:
        A JSON string indicating success or failure, including any warnings.
        Example success: '{"success": true, "path": "new_config.json"}'
        Example success with warning: '{"success": true, "path": "f.txt", "warning": "..."}'
        Example error: '{"success": false, "error": "Permission denied: /etc/hosts"}'
    """
    actual_lines = len(content.splitlines())
    warning = None
    if actual_lines != line_count:
        warning = (
            f"Provided line_count ({line_count}) does not match actual "
            f"content lines ({actual_lines}) for file {path}"
        )
    try:
        abs_path = os.path.abspath(path)
        # Ensure directory exists
        directory = os.path.dirname(abs_path)
        if directory and not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)
        _write_file(abs_path, content)
        result_data = {"success": True, "path": path}
        if warning:
            result_data["warning"] = warning
        return json.dumps(result_data)
    except Exception as e:
        raise Exception(f"Error writing file {e}")


def search_files(
    path: str,
    regex: str,
    file_pattern: Optional[str] = None,
    include_hidden: bool = False,
) -> str:
    """
    Request to perform a regex search across files in a specified directory,
    providing context-rich results. This tool searches for patterns or specific
    content across multiple files, displaying each match with encapsulating context.
    Args:
        path: (required) The path of the directory to search in (relative to the CWD).
              This directory will be recursively searched.
        regex: (required) The regular expression pattern to search for. Uses Rust regex syntax.
               (Note: Python's `re` module will be used here, which has similar syntax)
        file_pattern: (optional) Glob pattern to filter files (e.g., '*.ts').
                      If not provided, searches all files (*).
        include_hidden: (optional) Whether to include hidden files.
                      Defaults to False (exclude hidden files).
    Returns:
        A JSON string containing the search results or an error message.
        Example success: '{"summary": "Found 5 matches...", "results": [{"file":"f.py", ...}]}'
        Example no match: '{"summary": "No matches found...", "results": []}'
        Example error: '{"error": "Invalid regex: ..."}'
    """
    try:
        pattern = re.compile(regex)
    except re.error as e:
        raise Exception(f"Invalid regex pattern: {e}")
    search_results = {"summary": "", "results": []}
    match_count = 0
    searched_file_count = 0
    file_match_count = 0
    try:
        abs_path = os.path.abspath(path)
        for root, dirs, files in os.walk(abs_path):
            # Skip hidden directories
            dirs[:] = [d for d in dirs if include_hidden or not _is_hidden(d)]
            for filename in files:
                # Skip hidden files
                if not include_hidden and _is_hidden(filename):
                    continue
                # Apply file pattern filter if provided
                if file_pattern and not fnmatch.fnmatch(filename, file_pattern):
                    continue
                file_path = os.path.join(root, filename)
                rel_file_path = os.path.relpath(file_path, os.getcwd())
                searched_file_count += 1
                try:
                    matches = _get_file_matches(file_path, pattern)
                    if matches:
                        file_match_count += 1
                        match_count += len(matches)
                        search_results["results"].append(
                            {"file": rel_file_path, "matches": matches}
                        )
                except IOError as e:
                    search_results["results"].append(
                        {"file": rel_file_path, "error": str(e)}
                    )
        if match_count == 0:
            search_results["summary"] = (
                f"No matches found for pattern '{regex}' in path '{path}' "
                f"(searched {searched_file_count} files)."
            )
        else:
            search_results["summary"] = (
                f"Found {match_count} matches in {file_match_count} files "
                f"(searched {searched_file_count} files)."
            )
        return json.dumps(search_results, indent=2)  # Pretty print for readability
    except Exception as e:
        raise Exception(f"Error searching files: {e}")


def _get_file_matches(
    file_path: str, pattern: re.Pattern, context_lines: int = 2
) -> List[Dict[str, Any]]:
    """Search for regex matches in a file with context."""
    try:
        with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
            lines = f.readlines()
        matches = []
        for line_idx, line in enumerate(lines):
            if pattern.search(line):
                line_num = line_idx + 1
                context_start = max(0, line_idx - context_lines)
                context_end = min(len(lines), line_idx + context_lines + 1)
                match_data = {
                    "line_number": line_num,
                    "line_content": line.rstrip(),
                    "context_before": [
                        lines[j].rstrip() for j in range(context_start, line_idx)
                    ],
                    "context_after": [
                        lines[j].rstrip() for j in range(line_idx + 1, context_end)
                    ],
                }
                matches.append(match_data)
        return matches
    except Exception as e:
        raise IOError(f"Error reading {file_path}: {str(e)}")


def apply_diff(path: str, diff: str) -> str:
    """
    Request to replace existing code using a search and replace block.
    This tool allows for precise, surgical replaces to files by specifying exactly
    what content to search for and what to replace it with.
    The tool will maintain proper indentation and formatting while making changes.
    Only a single operation is allowed per tool use.
    The SEARCH section must exactly match existing content including whitespace
    and indentation.
    If you're not confident in the exact content to search for, use the read_file tool
    first to get the exact content.
    Args:
        path: (required) The path of the file to modify (relative to the CWD)
        diff: (required) The search/replace block defining the changes.
              Format:
              <<<<<<< SEARCH
              :start_line:START_LINE_NUMBER
              :end_line:END_LINE_NUMBER
              -------
              [exact content to find including whitespace]
              =======
              [new content to replace with]
              >>>>>>> REPLACE
    Returns:
        A JSON string indicating success or failure.
    """
    try:
        start_line, end_line, search_content, replace_content = _parse_diff(diff)
        abs_path = os.path.abspath(path)
        if not os.path.exists(abs_path):
            return json.dumps(
                {"success": False, "path": path, "error": f"File not found at {path}"}
            )
        content = _read_file(abs_path)
        lines = content.splitlines()
        if start_line < 1 or end_line > len(lines) or start_line > end_line:
            return json.dumps(
                {
                    "success": False,
                    "path": path,
                    "error": (
                        f"Invalid line range {start_line}-{end_line} "
                        f"for file with {len(lines)} lines."
                    ),
                }
            )
        original_content = "\n".join(lines[start_line - 1 : end_line])
        if original_content != search_content:
            error_message = (
                f"Search content does not match file content at "
                f"lines {start_line}-{end_line}.\n"
                f"Expected ({len(search_content.splitlines())} lines):\n"
                f"---\n{search_content}\n---\n"
                f"Actual ({len(lines[start_line-1:end_line])} lines):\n"
                f"---\n{original_content}\n---"
            )
            return json.dumps({"success": False, "path": path, "error": error_message})
        new_lines = (
            lines[: start_line - 1] + replace_content.splitlines() + lines[end_line:]
        )
        new_content = "\n".join(new_lines)
        if content.endswith("\n"):
            new_content += "\n"
        _write_file(abs_path, new_content)
        return json.dumps({"success": True, "path": path})
    except Exception as e:
        raise Exception(f"Error applying diff on {path}: {e}")


def _parse_diff(diff: str) -> tuple[int, int, str, str]:
    """Parse diff content into components."""
    search_marker = "<<<<<<< SEARCH"
    meta_marker = "-------"
    separator = "======="
    replace_marker = ">>>>>>> REPLACE"
    search_start_idx = diff.find(search_marker)
    meta_start_idx = diff.find(meta_marker)
    separator_idx = diff.find(separator)
    replace_end_idx = diff.find(replace_marker)
    if any(
        idx == -1
        for idx in [search_start_idx, meta_start_idx, separator_idx, replace_end_idx]
    ):
        raise ValueError("Invalid diff format - missing markers")
    meta_content = diff[search_start_idx + len(search_marker) : meta_start_idx].strip()
    start_line = int(re.search(r":start_line:(\d+)", meta_content).group(1))
    end_line = int(re.search(r":end_line:(\d+)", meta_content).group(1))
    search_content = diff[meta_start_idx + len(meta_marker) : separator_idx].strip(
        "\r\n"
    )
    replace_content = diff[separator_idx + len(separator) : replace_end_idx].strip(
        "\r\n"
    )
    return start_line, end_line, search_content, replace_content



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # We want fastapi to only be loaded when necessary to decrease footprint
    from fastapi import FastAPI


def serve_docs(app: "FastAPI") -> None:
    from fastapi.openapi.docs import get_swagger_ui_html

    @app.get("/docs", include_in_schema=False)
    async def swagger_ui_html():
        return get_swagger_ui_html(
            openapi_url="/openapi.json",
            title="Zrb",
            swagger_favicon_url="/static/favicon-32x32.png",
        )



import datetime

import ulid
from my_app_name.schema.permission import Permission
from my_app_name.schema.role import Role
from pydantic import BaseModel
from sqlmodel import Field, SQLModel


class UserBase(SQLModel):
    username: str
    active: bool


class UserCreate(UserBase):
    password: str

    def with_audit(self, created_by: str) -> "UserCreateWithAudit":
        return UserCreateWithAudit(**self.model_dump(), created_by=created_by)


class UserCreateWithAudit(UserCreate):
    created_by: str


class UserCreateWithRoles(UserCreate):
    role_names: list[str] | None = None

    def with_audit(self, created_by: str) -> "UserCreateWithRolesAndAudit":
        return UserCreateWithRolesAndAudit(**self.model_dump(), created_by=created_by)


class UserCreateWithRolesAndAudit(UserCreateWithRoles):
    created_by: str

    def get_user_create_with_audit(self) -> UserCreateWithAudit:
        data = {
            key: val for key, val in self.model_dump().items() if key != "role_names"
        }
        return UserCreateWithAudit(**data)

    def get_role_names(self) -> list[str]:
        if self.role_names is None:
            return []
        return self.role_names


class UserUpdate(SQLModel):
    username: str | None = None
    password: str | None = None
    active: bool | None = None

    def with_audit(self, updated_by: str) -> "UserUpdateWithAudit":
        return UserUpdateWithAudit(
            **self.model_dump(exclude_none=True), updated_by=updated_by
        )


class UserUpdateWithAudit(UserUpdate):
    updated_by: str


class UserUpdateWithRoles(UserUpdate):
    role_names: list[str] | None = None

    def with_audit(self, updated_by: str) -> "UserUpdateWithRolesAndAudit":
        return UserUpdateWithRolesAndAudit(
            **self.model_dump(exclude_none=True), updated_by=updated_by
        )


class UserUpdateWithRolesAndAudit(UserUpdateWithRoles):
    updated_by: str

    def get_user_update_with_audit(self) -> UserUpdateWithAudit:
        data = {
            key: val
            for key, val in self.model_dump(exclude_none=True).items()
            if key != "role_names"
        }
        return UserUpdateWithAudit(**data)

    def get_role_names(self) -> list[str]:
        if self.role_names is None:
            return []
        return self.role_names


class UserResponse(UserBase):
    id: str
    role_names: list[str]
    permission_names: list[str]


class AuthUserResponse(UserResponse):
    is_super_user: bool
    is_guest: bool

    def has_permission(self, permission_name: str):
        return self.is_super_user or permission_name in self.permission_names

    def has_role(self, role_name: str):
        return self.is_super_user or role_name in self.role_names


class MultipleUserResponse(BaseModel):
    data: list[UserResponse]
    count: int


class UserCredentials(SQLModel):
    username: str
    password: str


class UserTokenData(SQLModel):
    access_token: str
    refresh_token: str
    access_token_expired_at: datetime.datetime
    refresh_token_expired_at: datetime.datetime


class UserSessionResponse(SQLModel):
    id: str
    user_id: str
    access_token: str
    refresh_token: str
    token_type: str
    access_token_expired_at: datetime.datetime
    refresh_token_expired_at: datetime.datetime


class User(SQLModel, table=True):
    __tablename__ = "users"
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    created_at: datetime.datetime = Field(index=True)
    created_by: str = Field(index=True)
    updated_at: datetime.datetime | None = Field(index=True)
    updated_by: str | None = Field(index=True)
    username: str = Field(index=True, unique=True)
    password: str
    active: bool = Field(index=True)


class UserRole(SQLModel, table=True):
    __tablename__ = "user_roles"
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    user_id: str = Field(index=True)
    role_id: str = Field(index=True)
    created_at: datetime.datetime | None
    created_by: str | None


class UserSession(SQLModel, table=True):
    __tablename__ = "user_sessions"
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    user_id: str = Field(index=True)
    access_token: str = Field(index=True)
    refresh_token: str = Field(index=True)
    access_token_expired_at: datetime.datetime = Field(index=True)
    refresh_token_expired_at: datetime.datetime = Field(index=True)



from zrb.group.group import Group
from zrb.runner.cli import cli

base64_group = cli.add_group(Group(name="base64", description="📄 Base64 operations"))
random_group = cli.add_group(Group(name="random", description="🔀 Random operation"))
git_group = cli.add_group(Group(name="git", description="🌱 Git related commands"))
git_branch_group = git_group.add_group(
    Group(name="branch", description="🌿 Git branch related commands")
)
git_subtree_group = git_group.add_group(
    Group(name="subtree", description="🌳 Git subtree related commands")
)
llm_group = cli.add_group(Group(name="llm", description="🤖 LLM operations"))
md5_group = cli.add_group(Group(name="md5", description="🔢 Md5 operations"))
python_group = cli.add_group(
    Group(name="python", description="🐍 Python related commands")
)
todo_group = cli.add_group(Group(name="todo", description="✅ Todo management"))

shell_group = cli.add_group(
    Group(name="shell", description="💬 Shell related commands")
)
shell_autocomplete_group: Group = shell_group.add_group(
    Group(name="autocomplete", description="⌨️ Shell autocomplete related commands")
)

project_group = cli.add_group(
    Group(name="project", description="📁 Project related commands")
)
add_to_project_group = project_group.add_group(
    Group(name="add", description="➕ Add things to project")
)
add_fastapp_to_project_group = add_to_project_group.add_group(
    Group(name="fastapp", description="🚀 Add Fastapp resources")
)

setup_group = cli.add_group(Group(name="setup", description="🔧 Setup"))
setup_latex_group = setup_group.add_group(
    Group(name="latex", description="✍️ Setup LaTeX")
)



import asyncio
import os
from collections.abc import Callable, Coroutine
from typing import Any

from zrb.attr.type import BoolAttr, fstring
from zrb.context.any_context import AnyContext
from zrb.context.shared_context import AnySharedContext, SharedContext
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.session.any_session import AnySession
from zrb.session.session import Session
from zrb.task.any_task import AnyTask
from zrb.util.attr import get_bool_attr
from zrb.util.run import run_async
from zrb.xcom.xcom import Xcom


class BaseTask(AnyTask):
    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput | None] | AnyInput | None = None,
        env: list[AnyEnv | None] | AnyEnv | None = None,
        action: fstring | Callable[[AnyContext], Any] | None = None,
        execute_condition: BoolAttr = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
        successor: list[AnyTask] | AnyTask | None = None,
    ):
        self._name = name
        self._color = color
        self._icon = icon
        self._description = description
        self._cli_only = cli_only
        self._inputs = input
        self._envs = env
        self._retries = retries
        self._retry_period = retry_period
        self._upstreams = upstream
        self._fallbacks = fallback
        self._successors = successor
        self._readiness_checks = readiness_check
        self._readiness_check_delay = readiness_check_delay
        self._readiness_check_period = readiness_check_period
        self._readiness_failure_threshold = readiness_failure_threshold
        self._readiness_timeout = readiness_timeout
        self._monitor_readiness = monitor_readiness
        self._execute_condition = execute_condition
        self._action = action

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self._name}>"

    def __rshift__(self, other: AnyTask | list[AnyTask]) -> AnyTask | list[AnyTask]:
        try:
            if isinstance(other, AnyTask):
                other.append_upstream(self)
            elif isinstance(other, list):
                for task in other:
                    task.append_upstream(self)
            return other
        except Exception as e:
            raise ValueError(f"Invalid operation {self} >> {other}: {e}")

    def __lshift__(self, other: AnyTask | list[AnyTask]) -> AnyTask:
        try:
            self.append_upstream(other)
            return self
        except Exception as e:
            raise ValueError(f"Invalid operation {self} << {other}: {e}")

    @property
    def name(self) -> str:
        return self._name

    @property
    def color(self) -> int | None:
        return self._color

    @property
    def icon(self) -> str | None:
        return self._icon

    @property
    def description(self) -> str:
        return self._description if self._description is not None else self.name

    @property
    def cli_only(self) -> bool:
        return self._cli_only

    @property
    def envs(self) -> list[AnyEnv]:
        envs = []
        for upstream in self.upstreams:
            envs += upstream.envs
        if isinstance(self._envs, AnyEnv):
            envs.append(self._envs)
        elif self._envs is not None:
            envs += self._envs
        return [env for env in envs if env is not None]

    @property
    def inputs(self) -> list[AnyInput]:
        inputs = []
        for upstream in self.upstreams:
            self.__combine_inputs(inputs, upstream.inputs)
        if self._inputs is not None:
            self.__combine_inputs(inputs, self._inputs)
        return [task_input for task_input in inputs if task_input is not None]

    def __combine_inputs(
        self,
        inputs: list[AnyInput],
        other_inputs: list[AnyInput | None] | AnyInput | None,
    ):
        input_names = [task_input.name for task_input in inputs]
        if isinstance(other_inputs, AnyInput):
            other_inputs = [other_inputs]
        elif other_inputs is None:
            other_inputs = []
        for task_input in other_inputs:
            if task_input is None:
                continue
            if task_input.name not in input_names:
                inputs.append(task_input)

    @property
    def fallbacks(self) -> list[AnyTask]:
        if self._fallbacks is None:
            return []
        elif isinstance(self._fallbacks, AnyTask):
            return [self._fallbacks]
        return self._fallbacks

    def append_fallback(self, fallbacks: AnyTask | list[AnyTask]):
        fallback_list = [fallbacks] if isinstance(fallbacks, AnyTask) else fallbacks
        for fallback in fallback_list:
            self.__append_fallback(fallback)

    def __append_fallback(self, fallback: AnyTask):
        # Make sure self._fallbacks is a list
        if self._fallbacks is None:
            self._fallbacks = []
        elif isinstance(self._fallbacks, AnyTask):
            self._fallbacks = [self._fallbacks]
        # Add fallback if it was not on self._fallbacks
        if fallback not in self._fallbacks:
            self._fallbacks.append(fallback)

    @property
    def successors(self) -> list[AnyTask]:
        if self._successors is None:
            return []
        elif isinstance(self._successors, AnyTask):
            return [self._successors]
        return self._successors

    def append_successor(self, successors: AnyTask | list[AnyTask]):
        successor_list = [successors] if isinstance(successors, AnyTask) else successors
        for successor in successor_list:
            self.__append_successor(successor)

    def __append_successor(self, successor: AnyTask):
        # Make sure self._successors is a list
        if self._successors is None:
            self._successors = []
        elif isinstance(self._successors, AnyTask):
            self._successors = [self._successors]
        # Add successor if it was not on self._successors
        if successor not in self._successors:
            self._successors.append(successor)

    @property
    def readiness_checks(self) -> list[AnyTask]:
        if self._readiness_checks is None:
            return []
        elif isinstance(self._readiness_checks, AnyTask):
            return [self._readiness_checks]
        return self._readiness_checks

    def append_readiness_check(self, readiness_checks: AnyTask | list[AnyTask]):
        readiness_check_list = (
            [readiness_checks]
            if isinstance(readiness_checks, AnyTask)
            else readiness_checks
        )
        for readiness_check in readiness_check_list:
            self.__append_readiness_check(readiness_check)

    def __append_readiness_check(self, readiness_check: AnyTask):
        # Make sure self._readiness_checks is a list
        if self._readiness_checks is None:
            self._readiness_checks = []
        elif isinstance(self._readiness_checks, AnyTask):
            self._readiness_checks = [self._readiness_checks]
        # Add readiness_check if it was not on self._readiness_checks
        if readiness_check not in self._readiness_checks:
            self._readiness_checks.append(readiness_check)

    @property
    def upstreams(self) -> list[AnyTask]:
        if self._upstreams is None:
            return []
        elif isinstance(self._upstreams, AnyTask):
            return [self._upstreams]
        return self._upstreams

    def append_upstream(self, upstreams: AnyTask | list[AnyTask]):
        upstream_list = [upstreams] if isinstance(upstreams, AnyTask) else upstreams
        for upstream in upstream_list:
            self.__append_upstream(upstream)

    def __append_upstream(self, upstream: AnyTask):
        # Make sure self._upstreams is a list
        if self._upstreams is None:
            self._upstreams = []
        elif isinstance(self._upstreams, AnyTask):
            self._upstreams = [self._upstreams]
        # Add upstream if it was not on self._upstreams
        if upstream not in self._upstreams:
            self._upstreams.append(upstream)

    def get_ctx(self, session: AnySession) -> AnyContext:
        ctx = session.get_ctx(self)
        # Enhance session ctx with current task env
        for env in self.envs:
            env.update_context(ctx)
        return ctx

    def run(
        self, session: AnySession | None = None, str_kwargs: dict[str, str] = {}
    ) -> Any:
        return asyncio.run(self._run_and_cleanup(session, str_kwargs))

    async def _run_and_cleanup(
        self,
        session: AnySession | None = None,
        str_kwargs: dict[str, str] = {},
    ) -> Any:
        current_task = asyncio.create_task(self.async_run(session, str_kwargs))
        try:
            result = await current_task
        finally:
            if session and not session.is_terminated:
                session.terminate()
            # Cancel all running tasks except the current one
            pending = [task for task in asyncio.all_tasks() if task is not current_task]
            for task in pending:
                task.cancel()
            if pending:
                try:
                    await asyncio.wait(pending, timeout=5)
                except asyncio.CancelledError:
                    pass
        return result

    async def async_run(
        self, session: AnySession | None = None, str_kwargs: dict[str, str] = {}
    ) -> Any:
        if session is None:
            session = Session(shared_ctx=SharedContext())
        # Update session
        self.__fill_shared_context_inputs(session.shared_ctx, str_kwargs)
        self.__fill_shared_context_envs(session.shared_ctx)
        result = await run_async(self.exec_root_tasks(session))
        return result

    def __fill_shared_context_inputs(
        self, shared_context: AnySharedContext, str_kwargs: dict[str, str] = {}
    ):
        for task_input in self.inputs:
            if task_input.name not in shared_context.input:
                str_value = str_kwargs.get(task_input.name, None)
                task_input.update_shared_context(shared_context, str_value)

    def __fill_shared_context_envs(self, shared_context: AnySharedContext):
        # Inject os environ
        os_env_map = {
            key: val for key, val in os.environ.items() if key not in shared_context.env
        }
        shared_context.env.update(os_env_map)

    async def exec_root_tasks(self, session: AnySession):
        session.set_main_task(self)
        session.state_logger.write(session.as_state_log())
        try:
            log_state = asyncio.create_task(self._log_session_state(session))
            root_tasks = [
                task
                for task in session.get_root_tasks(self)
                if session.is_allowed_to_run(task)
            ]
            root_task_coros = [
                run_async(root_task.exec_chain(session)) for root_task in root_tasks
            ]
            await asyncio.gather(*root_task_coros)
            await session.wait_deferred()
            session.terminate()
            await log_state
            return session.final_result
        except IndexError:
            return None
        except (asyncio.CancelledError, KeyboardInterrupt):
            ctx = self.get_ctx(session)
            ctx.log_info("Session terminated")
        finally:
            session.terminate()
            session.state_logger.write(session.as_state_log())
            ctx = self.get_ctx(session)
            ctx.log_debug(session)

    async def _log_session_state(self, session: AnySession):
        session.state_logger.write(session.as_state_log())
        while not session.is_terminated:
            await asyncio.sleep(0.1)
            session.state_logger.write(session.as_state_log())
        session.state_logger.write(session.as_state_log())

    async def exec_chain(self, session: AnySession):
        if session.is_terminated or not session.is_allowed_to_run(self):
            return
        result = await self.exec(session)
        # Get next tasks
        nexts = session.get_next_tasks(self)
        if session.is_terminated or len(nexts) == 0:
            return result
        # Run next tasks asynchronously
        next_coros = [run_async(next.exec_chain(session)) for next in nexts]
        return await asyncio.gather(*next_coros)

    async def exec(self, session: AnySession):
        ctx = self.get_ctx(session)
        if not session.is_allowed_to_run(self):
            # Task is not allowed to run, skip it for now.
            # This will be triggered later
            ctx.log_info("Not allowed to run")
            return
        if not self.__get_execute_condition(session):
            # Skip the task
            ctx.log_info("Marked as skipped")
            session.get_task_status(self).mark_as_skipped()
            return
        # Wait for task to be ready
        return await run_async(self.__exec_action_until_ready(session))

    def __get_execute_condition(self, session: Session) -> bool:
        ctx = self.get_ctx(session)
        return get_bool_attr(ctx, self._execute_condition, True, auto_render=True)

    async def __exec_action_until_ready(self, session: AnySession):
        ctx = self.get_ctx(session)
        readiness_checks = self.readiness_checks
        if len(readiness_checks) == 0:
            ctx.log_info("No readiness checks")
            # Task has no readiness check
            result = await run_async(self.__exec_action_and_retry(session))
            ctx.log_info("Marked as ready")
            session.get_task_status(self).mark_as_ready()
            return result
        # Start the task along with the readiness checks
        action_coro = asyncio.create_task(
            run_async(self.__exec_action_and_retry(session))
        )
        await asyncio.sleep(self._readiness_check_delay)
        readiness_check_coros = [
            run_async(check.exec_chain(session)) for check in readiness_checks
        ]
        # Only wait for readiness checks and mark the task as ready
        ctx.log_info("Start readiness checks")
        result = await asyncio.gather(*readiness_check_coros)
        ctx.log_info("Readiness checks completed")
        ctx.log_info("Marked as ready")
        session.get_task_status(self).mark_as_ready()
        # Defer task's coroutines, will be waited later
        session.defer_action(self, action_coro)
        if self._monitor_readiness:
            monitor_and_rerun_coro = asyncio.create_task(
                run_async(self.__exec_monitoring(session, action_coro))
            )
            session.defer_monitoring(self, monitor_and_rerun_coro)
        return result

    async def __exec_monitoring(self, session: AnySession, action_coro: asyncio.Task):
        readiness_checks = self.readiness_checks
        failure_count = 0
        ctx = self.get_ctx(session)
        while not session.is_terminated:
            await asyncio.sleep(self._readiness_check_period)
            if failure_count < self._readiness_failure_threshold:
                for readiness_check in readiness_checks:
                    session.get_task_status(readiness_check).reset_history()
                    session.get_task_status(readiness_check).reset()
                    readiness_xcom: Xcom = ctx.xcom[self.name]
                    readiness_xcom.clear()
                readiness_check_coros = [
                    check.exec_chain(session) for check in readiness_checks
                ]
                try:
                    ctx.log_info("Checking")
                    await asyncio.wait_for(
                        asyncio.gather(*readiness_check_coros),
                        timeout=self._readiness_timeout,
                    )
                    ctx.log_info("OK")
                    continue
                except (asyncio.CancelledError, KeyboardInterrupt):
                    for readiness_check in readiness_checks:
                        ctx.log_info("Marked as failed")
                        session.get_task_status(readiness_check).mark_as_failed()
                except asyncio.TimeoutError:
                    failure_count += 1
                    ctx.log_info("Detecting failure")
                    ctx.log_debug(f"Failure count: {failure_count}")
            # Readiness check failed, reset
            ctx.log_info("Resetting")
            action_coro.cancel()
            session.get_task_status(self).reset()
            # defer this action
            ctx.log_info("Running")
            action_coro: Coroutine = asyncio.create_task(
                run_async(self.__exec_action_and_retry(session))
            )
            session.defer_action(self, action_coro)
            failure_count = 0
            ctx.log_info("Continue monitoring")

    async def __exec_action_and_retry(self, session: AnySession) -> Any:
        """
        Executes an action with retry logic.

        This method attempts to execute the action defined in `_exec_action` with a specified number of retries.
        If the action fails, it will retry after a specified period until the maximum number of attempts is reached.
        If the action succeeds, it marks the task as completed and executes any successors.
        If the action fails permanently, it marks the task as permanently failed and executes any fallbacks.

        Args:
            session (AnySession): The session object containing the task status and context.

        Returns:
            Any: The result of the executed action if successful.

        Raises:
            Exception: If the action fails permanently after all retry attempts.
        """
        ctx = self.get_ctx(session)
        max_attempt = self._retries + 1
        ctx.set_max_attempt(max_attempt)
        for attempt in range(max_attempt):
            ctx.set_attempt(attempt + 1)
            if attempt > 0:
                # apply retry period only if this is not the first attempt
                await asyncio.sleep(self._retry_period)
            try:
                ctx.log_info("Marked as started")
                session.get_task_status(self).mark_as_started()
                result = await run_async(self._exec_action(ctx))
                ctx.log_info("Marked as completed")
                session.get_task_status(self).mark_as_completed()
                # Put result on xcom
                task_xcom: Xcom = ctx.xcom.get(self.name)
                task_xcom.push(result)
                self.__skip_fallbacks(session)
                await run_async(self.__exec_successors(session))
                return result
            except (asyncio.CancelledError, KeyboardInterrupt):
                ctx.log_info("Marked as failed")
                session.get_task_status(self).mark_as_failed()
                return
            except BaseException as e:
                ctx.log_error(e)
                if attempt < max_attempt - 1:
                    ctx.log_info("Marked as failed")
                    session.get_task_status(self).mark_as_failed()
                    continue
                ctx.log_info("Marked as permanently failed")
                session.get_task_status(self).mark_as_permanently_failed()
                self.__skip_successors(session)
                await run_async(self.__exec_fallbacks(session))
                raise e

    async def __exec_successors(self, session: AnySession) -> Any:
        successors: list[AnyTask] = self.successors
        successor_coros = [
            run_async(successor.exec_chain(session)) for successor in successors
        ]
        await asyncio.gather(*successor_coros)

    def __skip_successors(self, session: AnySession) -> Any:
        for successor in self.successors:
            session.get_task_status(successor).mark_as_skipped()

    async def __exec_fallbacks(self, session: AnySession) -> Any:
        fallbacks: list[AnyTask] = self.fallbacks
        fallback_coros = [
            run_async(fallback.exec_chain(session)) for fallback in fallbacks
        ]
        await asyncio.gather(*fallback_coros)

    def __skip_fallbacks(self, session: AnySession) -> Any:
        for fallback in self.fallbacks:
            session.get_task_status(fallback).mark_as_skipped()

    async def _exec_action(self, ctx: AnyContext) -> Any:
        """Execute the main action of the task.
        By default will render and run the _action attribute.

        Override this method to define custom action.

        Args:
            session (AnySession): The shared session.

        Returns:
            Any: The result of the action execution.
        """
        if self._action is None:
            return
        if isinstance(self._action, str):
            return ctx.render(self._action)
        return await run_async(self._action(ctx))



import json
import os
from typing import Any

from zrb.builtin.group import llm_group
from zrb.builtin.llm.tool.api import get_current_location, get_current_weather
from zrb.builtin.llm.tool.cli import run_shell_command
from zrb.builtin.llm.tool.file import (
    apply_diff,
    list_files,
    read_from_file,
    search_files,
    write_to_file,
)
from zrb.builtin.llm.tool.web import (
    create_search_internet_tool,
    open_web_page,
    search_arxiv,
    search_wikipedia,
)
from zrb.config import (
    LLM_ALLOW_ACCESS_INTERNET,
    LLM_ALLOW_ACCESS_LOCAL_FILE,
    LLM_ALLOW_ACCESS_SHELL,
    LLM_HISTORY_DIR,
    SERP_API_KEY,
)
from zrb.context.any_shared_context import AnySharedContext
from zrb.input.bool_input import BoolInput
from zrb.input.str_input import StrInput
from zrb.input.text_input import TextInput
from zrb.task.llm_task import LLMTask
from zrb.util.file import read_file, write_file
from zrb.util.string.conversion import to_pascal_case


class PreviousSessionInput(StrInput):

    def to_html(self, ctx: AnySharedContext) -> str:
        name = self.name
        description = self.description
        default = self.get_default_str(ctx)
        script = read_file(
            file_path=os.path.join(os.path.dirname(__file__), "previous-session.js"),
            replace_map={
                "CURRENT_INPUT_NAME": name,
                "CurrentPascalInputName": to_pascal_case(name),
            },
        )
        return "\n".join(
            [
                f'<input name="{name}" placeholder="{description}" value="{default}" />',
                f"<script>{script}</script>",
            ]
        )


def _read_chat_conversation(ctx: AnySharedContext) -> list[dict[str, Any]]:
    if ctx.input.start_new:
        return []
    previous_session_name = ctx.input.previous_session
    if previous_session_name == "" or previous_session_name is None:
        last_session_file_path = os.path.join(LLM_HISTORY_DIR, "last-session")
        if os.path.isfile(last_session_file_path):
            previous_session_name = read_file(last_session_file_path).strip()
    conversation_file_path = os.path.join(
        LLM_HISTORY_DIR, f"{previous_session_name}.json"
    )
    if not os.path.isfile(conversation_file_path):
        return []
    return json.loads(read_file(conversation_file_path))


def _write_chat_conversation(
    ctx: AnySharedContext, conversations: list[dict[str, Any]]
):
    os.makedirs(LLM_HISTORY_DIR, exist_ok=True)
    current_session_name = ctx.session.name
    conversation_file_path = os.path.join(
        LLM_HISTORY_DIR, f"{current_session_name}.json"
    )
    write_file(conversation_file_path, json.dumps(conversations, indent=2))
    last_session_file_path = os.path.join(LLM_HISTORY_DIR, "last-session")
    write_file(last_session_file_path, current_session_name)


llm_chat: LLMTask = llm_group.add_task(
    LLMTask(
        name="llm-chat",
        input=[
            StrInput(
                "model",
                description="LLM Model",
                prompt="LLM Model",
                default="",
                allow_positional_parsing=False,
                always_prompt=False,
                allow_empty=True,
            ),
            StrInput(
                "base-url",
                description="LLM API Base URL",
                prompt="LLM API Base URL",
                default="",
                allow_positional_parsing=False,
                always_prompt=False,
                allow_empty=True,
            ),
            StrInput(
                "api-key",
                description="LLM API Key",
                prompt="LLM API Key",
                default="",
                allow_positional_parsing=False,
                always_prompt=False,
                allow_empty=True,
            ),
            TextInput(
                "system-prompt",
                description="System prompt",
                prompt="System prompt",
                default="",
                allow_positional_parsing=False,
                always_prompt=False,
            ),
            BoolInput(
                "start-new",
                description="Start new conversation (LLM will forget everything)",
                prompt="Start new conversation (LLM will forget everything)",
                default=False,
                allow_positional_parsing=False,
                always_prompt=False,
            ),
            TextInput("message", description="User message", prompt="Your message"),
            PreviousSessionInput(
                "previous-session",
                description="Previous conversation session",
                prompt="Previous conversation session (can be empty)",
                allow_positional_parsing=False,
                allow_empty=True,
                always_prompt=False,
            ),
        ],
        model=lambda ctx: None if ctx.input.model.strip() == "" else ctx.input.model,
        model_base_url=lambda ctx: (
            None if ctx.input.base_url.strip() == "" else ctx.input.base_url
        ),
        model_api_key=lambda ctx: (
            None if ctx.input.api_key.strip() == "" else ctx.input.api_key
        ),
        conversation_history_reader=_read_chat_conversation,
        conversation_history_writer=_write_chat_conversation,
        description="Chat with LLM",
        system_prompt=lambda ctx: (
            None if ctx.input.system_prompt.strip() == "" else ctx.input.system_prompt
        ),
        message="{ctx.input.message}",
        retries=0,
    ),
    alias="chat",
)


if LLM_ALLOW_ACCESS_LOCAL_FILE:
    llm_chat.add_tool(list_files)
    llm_chat.add_tool(read_from_file)
    llm_chat.add_tool(write_to_file)
    llm_chat.add_tool(search_files)
    llm_chat.add_tool(apply_diff)

if LLM_ALLOW_ACCESS_SHELL:
    llm_chat.add_tool(run_shell_command)

if LLM_ALLOW_ACCESS_INTERNET:
    llm_chat.add_tool(open_web_page)
    llm_chat.add_tool(search_wikipedia)
    llm_chat.add_tool(search_arxiv)
    if SERP_API_KEY != "":
        llm_chat.add_tool(create_search_internet_tool(SERP_API_KEY))
    llm_chat.add_tool(get_current_location)
    llm_chat.add_tool(get_current_weather)



import os

from zrb.group.any_group import AnyGroup
from zrb.runner.web_schema.user import User
from zrb.runner.web_util.html import (
    get_html_auth_link,
    get_html_subgroup_info,
    get_html_subtask_info,
)
from zrb.util.file import read_file
from zrb.util.string.format import fstring_format


def show_group_page(user: User, root_group: AnyGroup, group: AnyGroup, url: str):
    from fastapi.responses import HTMLResponse

    _DIR = os.path.dirname(__file__)
    _VIEW_TEMPLATE = read_file(os.path.join(_DIR, "view.html"))
    url_parts = url.split("/")
    parent_url_parts = url_parts[:-2] + [""]
    parent_url = "/".join(parent_url_parts)
    group_info = get_html_subgroup_info(user, url, group)
    task_info = get_html_subtask_info(user, url, group)
    auth_link = get_html_auth_link(user)
    return HTMLResponse(
        fstring_format(
            _VIEW_TEMPLATE,
            {
                "group_info": group_info,
                "task_info": task_info,
                "name": group.name,
                "description": group.description,
                "auth_link": auth_link,
                "root_name": root_group.name,
                "root_description": root_group.description,
                "url": url,
                "parent_url": parent_url,
            },
        )
    )



import libcst as cst

from zrb.util.codemod.modification_mode import APPEND, PREPEND


def prepend_key_to_dict(
    original_code: str, dictionary_name: str, new_key: str, new_value: str
) -> str:
    return _modify_dict(original_code, dictionary_name, new_key, new_value, PREPEND)


def append_key_to_dict(
    original_code: str, dictionary_name: str, new_key: str, new_value: str
) -> str:
    return _modify_dict(original_code, dictionary_name, new_key, new_value, APPEND)


def _modify_dict(
    original_code: str, dictionary_name: str, new_key: str, new_value: str, mode: int
) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize the transformer with the necessary information
    transformer = _DictionaryModifier(dictionary_name, new_key, new_value, mode)
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Error handling: raise an error if the dictionary is not found
    if not transformer.found:
        raise ValueError(
            f"Dictionary {dictionary_name} not found in the provided code."
        )
    # Return the modified code
    return modified_module.code


class _DictionaryModifier(cst.CSTTransformer):
    def __init__(self, dictionary_name: str, new_key: str, new_value: str, mode: int):
        self.dictionary_name = dictionary_name
        self.new_key = new_key
        self.new_value = new_value
        self.found = False
        self.mode = mode

    def leave_Assign(
        self, original_node: cst.Assign, updated_node: cst.Assign
    ) -> cst.Assign:
        # Extract the first target from updated_node, which will be an AssignTarget
        target = updated_node.targets[0]
        # Check if the target is a Name (which should represent the dictionary)
        if (
            isinstance(target.target, cst.Name)
            and target.target.value == self.dictionary_name
        ):
            # Check if it's a dictionary initialization (e.g., my_dict = {...})
            if isinstance(updated_node.value, cst.Dict):
                self.found = True
                if self.mode == PREPEND:
                    new_entries = (
                        cst.DictElement(
                            key=cst.SimpleString(f'"{self.new_key}"'),
                            value=cst.SimpleString(f'"{self.new_value}"'),
                        ),
                    ) + updated_node.value.elements
                    new_dict = updated_node.value.with_changes(elements=new_entries)
                    return updated_node.with_changes(value=new_dict)
                if self.mode == APPEND:
                    new_entries = updated_node.value.elements + (
                        cst.DictElement(
                            key=cst.SimpleString(f'"{self.new_key}"'),
                            value=cst.SimpleString(f'"{self.new_value}"'),
                        ),
                    )
                    new_dict = updated_node.value.with_changes(elements=new_entries)
                    return updated_node.with_changes(value=new_dict)
        return updated_node



import re

BLACK = 30
RED = 31
GREEN = 32
YELLOW = 33
BLUE = 34
MAGENTA = 35
CYAN = 36
WHITE = 37
BRIGHT_BLACK = 90
BRIGHT_RED = 91
BRIGHT_GREEN = 92
BRIGHT_YELLOW = 93
BRIGHT_BLUE = 94
BRIGHT_MAGENTA = 95
BRIGHT_CYAN = 96
BRIGHT_WHITE = 97

VALID_COLORS = [
    BLACK,
    RED,
    GREEN,
    YELLOW,
    BLUE,
    MAGENTA,
    CYAN,
    WHITE,
    BRIGHT_BLACK,
    BRIGHT_RED,
    BRIGHT_GREEN,
    BRIGHT_YELLOW,
    BRIGHT_BLUE,
    BRIGHT_MAGENTA,
    BRIGHT_CYAN,
    BRIGHT_WHITE,
]

BG_BLACK = 40
BG_RED = 41
BG_GREEN = 42
BG_YELLOW = 43
BG_BLUE = 44
BG_MAGENTA = 45
BG_CYAN = 46
BG_WHITE = 47
BG_BRIGHT_BLACK = 100
BG_BRIGHT_RED = 101
BG_BRIGHT_GREEN = 102
BG_BRIGHT_YELLOW = 103
BG_BRIGHT_BLUE = 104
BG_BRIGHT_MAGENTA = 105
BG_BRIGHT_CYAN = 106
BG_BRIGHT_WHITE = 107

VALID_BACKGROUNDS = [
    BG_BLACK,
    BG_RED,
    BG_GREEN,
    BG_YELLOW,
    BG_BLUE,
    BG_MAGENTA,
    BG_CYAN,
    BG_WHITE,
    BG_BRIGHT_BLACK,
    BG_BRIGHT_RED,
    BG_BRIGHT_GREEN,
    BG_BRIGHT_YELLOW,
    BG_BRIGHT_BLUE,
    BG_BRIGHT_MAGENTA,
    BG_BRIGHT_CYAN,
    BG_BRIGHT_WHITE,
]

BOLD = 1  # Bold or increased intensity
FAINT = 2  # Faint, decreased intensity, or dim
ITALIC = 3  # Italicized (not widely supported)
UNDERLINE = 4  # Underline
BLINK_SLOW = 5  # Slow blink (less than 150 per minute)
BLINK_FAST = 6  # Rapid blink (MS-DOS style, not widely supported)
REVERSED = 7  # Image negative, swap foreground and background colors
HIDE = 8  # Conceal (not widely supported)
CROSSED_OUT = 9  # Crossed-out or strike-through text

VALID_STYLES = [
    BOLD,
    FAINT,
    ITALIC,
    UNDERLINE,
    BLINK_SLOW,
    BLINK_FAST,
    REVERSED,
    HIDE,
    CROSSED_OUT,
]

ICONS = [
    "🐶",
    "🐱",
    "🐭",
    "🐹",
    "🦊",
    "🐻",
    "🐨",
    "🐯",
    "🦁",
    "🐮",
    "🐷",
    "🍎",
    "🍐",
    "🍊",
    "🍋",
    "🍌",
    "🍉",
    "🍇",
    "🍓",
    "🍈",
    "🍒",
    "🍑",
]


def remove_style(text):
    ansi_escape = re.compile(r"\x1B[@-_][0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", text)


def stylize(
    text: str,
    color: int | None = None,
    background: int | None = None,
    style: int | None = None,
):
    # Start constructing the ANSI escape code
    code_parts = []
    if style is not None and style in VALID_STYLES:
        code_parts.append(str(style))
    if color is not None and color in VALID_COLORS:
        code_parts.append(str(color))
    if background is not None and background in VALID_BACKGROUNDS:
        code_parts.append(str(background))
    # Join all parts with ';' and add the escape code ending
    if len(code_parts) > 0:
        return "\033[" + ";".join(code_parts) + "m" + text + "\033[0m"
    return text


def stylize_section_header(text: str):
    return stylize(f" {text} ", color=BLACK, background=BG_WHITE, style=UNDERLINE)


def stylize_green(text: str):
    return stylize(text, color=GREEN)


def stylize_blue(text: str):
    return stylize(text, color=BLUE)


def stylize_cyan(text: str):
    return stylize(text, color=CYAN)


def stylize_magenta(text: str):
    return stylize(text, color=MAGENTA)


def stylize_yellow(text: str):
    return stylize(text, color=YELLOW)


def stylize_red(text: str):
    return stylize(text, color=RED)


def stylize_bold_green(text: str):
    return stylize(text, color=GREEN, style=BOLD)


def stylize_bold_yellow(text: str):
    return stylize(text, color=YELLOW, style=BOLD)


def stylize_bold_red(text: str):
    return stylize(text, color=RED, style=BOLD)


def stylize_faint(text: str):
    return stylize(text, style=FAINT)


def stylize_log(text: str):
    return stylize_faint(text)


def stylize_warning(text: str):
    return stylize_bold_yellow(text)


def stylize_error(text: str):
    return stylize_bold_red(text)



import time
from typing import Annotated

from fastapi import Depends
from fastapi.testclient import TestClient
from my_app_name.config import (
    APP_AUTH_ACCESS_TOKEN_COOKIE_NAME,
    APP_AUTH_GUEST_USER,
    APP_AUTH_GUEST_USER_PERMISSIONS,
    APP_AUTH_REFRESH_TOKEN_COOKIE_NAME,
    APP_AUTH_SUPER_USER,
    APP_AUTH_SUPER_USER_PASSWORD,
)
from my_app_name.main import app
from my_app_name.module.gateway.util.auth import get_current_user
from my_app_name.schema.user import AuthUserResponse


@app.get("/test/current-user", response_model=AuthUserResponse)
async def serve_get_current_user(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
) -> AuthUserResponse:
    return current_user


def test_create_invalid_user_session():
    client = TestClient(app, base_url="http://localhost")
    response = client.post(
        "/api/v1/user-sessions",
        data={
            "username": "nonExistingUserFromSevenKingdom",
            "password": "nonExistingSecurity",
        },
    )
    assert response.status_code == 401


def test_get_guest_user():
    client = TestClient(app, base_url="http://localhost")
    # Fetch current user
    response = client.get("/test/current-user")
    assert response.status_code == 200
    user_data = response.json()
    assert user_data.get("username") == APP_AUTH_GUEST_USER
    assert not user_data.get("is_super_user")
    assert user_data.get("is_guest")
    assert user_data.get("permission_names") == APP_AUTH_GUEST_USER_PERMISSIONS


def test_create_admin_user_session():
    client = TestClient(app, base_url="http://localhost")
    # Create new admin user session and check the response
    session_response = client.post(
        "/api/v1/user-sessions",
        data={
            "username": APP_AUTH_SUPER_USER,
            "password": APP_AUTH_SUPER_USER_PASSWORD,
        },
    )
    assert session_response.status_code == 200
    session_data = session_response.json()
    assert session_data.get("user_id") == APP_AUTH_SUPER_USER
    assert session_data.get("token_type") == "bearer"
    assert "access_token" in session_data
    assert "access_token_expired_at" in session_data
    assert "refresh_token" in session_data
    assert "refresh_token_expired_at" in session_data
    assert session_response.cookies.get(
        APP_AUTH_ACCESS_TOKEN_COOKIE_NAME
    ) == session_data.get("access_token")
    assert session_response.cookies.get(
        APP_AUTH_REFRESH_TOKEN_COOKIE_NAME
    ) == session_data.get("refresh_token")


def test_get_admin_user_with_bearer():
    client = TestClient(app, base_url="http://localhost")
    session_response = client.post(
        "/api/v1/user-sessions",
        data={
            "username": APP_AUTH_SUPER_USER,
            "password": APP_AUTH_SUPER_USER_PASSWORD,
        },
    )
    assert session_response.status_code == 200
    access_token = session_response.json()["access_token"]
    # Fetch current user with bearer token
    response = client.get(
        "/test/current-user", headers={"Authorization": f"Bearer {access_token}"}
    )
    assert response.status_code == 200
    user_data = response.json()
    assert user_data.get("username") == APP_AUTH_SUPER_USER
    assert user_data.get("is_super_user")
    assert not user_data.get("is_guest")


def test_get_admin_user_with_cookie():
    login_client = TestClient(app, base_url="http://localhost")
    session_response = login_client.post(
        "/api/v1/user-sessions",
        data={
            "username": APP_AUTH_SUPER_USER,
            "password": APP_AUTH_SUPER_USER_PASSWORD,
        },
    )
    assert session_response.status_code == 200
    # Fetch current user with cookies
    cookies = {
        APP_AUTH_ACCESS_TOKEN_COOKIE_NAME: session_response.cookies.get(
            APP_AUTH_ACCESS_TOKEN_COOKIE_NAME
        )
    }
    # re-initiate client with cookies
    client = TestClient(app, base_url="http://localhost", cookies=cookies)
    response = client.get("/test/current-user")
    assert response.status_code == 200
    user_data = response.json()
    assert user_data.get("username") == APP_AUTH_SUPER_USER
    assert user_data.get("is_super_user")
    assert not user_data.get("is_guest")


def test_update_user_session():
    login_client = TestClient(app, base_url="http://localhost")
    old_session_response = login_client.post(
        "/api/v1/user-sessions",
        data={
            "username": APP_AUTH_SUPER_USER,
            "password": APP_AUTH_SUPER_USER_PASSWORD,
        },
    )
    assert old_session_response.status_code == 200
    old_session_data = old_session_response.json()
    # re-initiate client with cookies and delete user session
    client = TestClient(app, base_url="http://localhost")
    time.sleep(1)
    new_session_response = client.put(
        "/api/v1/user-sessions",
        params={"refresh_token": old_session_data.get("refresh_token")},
    )
    assert new_session_response.status_code == 200
    new_session_data = new_session_response.json()
    # Cookies and response should match
    assert new_session_response.cookies.get(
        APP_AUTH_ACCESS_TOKEN_COOKIE_NAME
    ) == new_session_data.get("access_token")
    assert new_session_response.cookies.get(
        APP_AUTH_REFRESH_TOKEN_COOKIE_NAME
    ) == new_session_data.get("refresh_token")
    # New session should has longer TTL than old session
    assert old_session_data.get("access_token") != new_session_data.get("access_token")
    assert old_session_data.get("access_token_expired_at") < new_session_data.get(
        "access_token_expired_at"
    )
    assert old_session_data.get("refresh_token") != new_session_data.get(
        "refresh_token"
    )
    assert old_session_data.get("refresh_token_expired_at") < new_session_data.get(
        "refresh_token_expired_at"
    )


def test_delete_user_session():
    login_client = TestClient(app, base_url="http://localhost")
    session_response = login_client.post(
        "/api/v1/user-sessions",
        data={
            "username": APP_AUTH_SUPER_USER,
            "password": APP_AUTH_SUPER_USER_PASSWORD,
        },
    )
    assert session_response.status_code == 200
    # Initiate cookies that should be deleted when user session is deleted.
    cookies = {
        APP_AUTH_ACCESS_TOKEN_COOKIE_NAME: session_response.cookies.get(
            APP_AUTH_ACCESS_TOKEN_COOKIE_NAME
        ),
        APP_AUTH_REFRESH_TOKEN_COOKIE_NAME: session_response.cookies.get(
            APP_AUTH_REFRESH_TOKEN_COOKIE_NAME
        ),
    }
    # re-initiate client with cookies and delete user session
    client = TestClient(app, base_url="http://localhost", cookies=cookies)
    response = client.delete(
        "/api/v1/user-sessions",
        params={
            "refresh_token": session_response.cookies.get(
                APP_AUTH_REFRESH_TOKEN_COOKIE_NAME
            ),
        },
    )
    assert response.status_code == 200
    assert response.cookies.get(APP_AUTH_ACCESS_TOKEN_COOKIE_NAME, None) is None
    assert response.cookies.get(APP_AUTH_REFRESH_TOKEN_COOKIE_NAME, None) is None



import os
from collections.abc import Callable
from typing import Any

from pydantic import BaseModel

from zrb.util.cmd.command import run_command
from zrb.util.file import read_file, write_file


class SingleSubTreeConfig(BaseModel):
    repo_url: str
    branch: str
    prefix: str


class SubTreeConfig(BaseModel):
    data: dict[str, SingleSubTreeConfig]


def load_config(repo_dir: str) -> SubTreeConfig:
    file_path = os.path.join(repo_dir, "subtrees.json")
    if not os.path.exists(file_path):
        return SubTreeConfig(data={})
    return SubTreeConfig.model_validate_json(read_file(file_path))


def save_config(repo_dir: str, config: SubTreeConfig):
    file_path = os.path.join(repo_dir, "subtrees.json")
    write_file(file_path, config.model_dump_json(indent=2))


async def add_subtree(
    repo_dir: str,
    name: str,
    repo_url: str,
    branch: str,
    prefix: str,
    print_method: Callable[..., Any] = print,
):
    config = load_config()
    if os.path.isdir(prefix):
        raise ValueError(f"Directory exists: {prefix}")
    if name in config.data:
        raise ValueError(f"Subtree config already exists: {name}")
    _, exit_code = await run_command(
        cmd=[
            "git",
            "subtree",
            "add",
            "--prefix",
            prefix,
            repo_url,
            branch,
        ],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    config.data[name] = SingleSubTreeConfig(
        repo_url=repo_url, branch=branch, prefix=prefix
    )
    save_config(repo_dir, config)


async def pull_subtree(
    repo_dir: str,
    prefix: str,
    repo_url: str,
    branch: str,
    print_method: Callable[..., Any] = print,
):
    _, exit_code = await run_command(
        cmd=[
            "git",
            "subtree",
            "pull",
            "--prefix",
            prefix,
            repo_url,
            branch,
        ],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")


async def push_subtree(
    repo_dir: str,
    prefix: str,
    repo_url: str,
    branch: str,
    print_method: Callable[..., Any] = print,
):
    _, exit_code = await run_command(
        cmd=[
            "git",
            "subtree",
            "push",
            "--prefix",
            prefix,
            repo_url,
            branch,
        ],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")



import os

from zrb.builtin.group import setup_group
from zrb.builtin.setup.common_input import package_manager_input, use_sudo_input
from zrb.builtin.setup.tmux.tmux_helper import get_install_tmux_cmd
from zrb.context.any_context import AnyContext
from zrb.input.str_input import StrInput
from zrb.task.cmd_task import CmdTask
from zrb.task.make_task import make_task
from zrb.util.file import read_file, write_file

install_tmux = CmdTask(
    name="install-tmux",
    input=[package_manager_input, use_sudo_input],
    cmd=get_install_tmux_cmd,
)


@make_task(
    name="setup-tmux",
    input=StrInput(
        name="tmux-config",
        description="Tmux config file",
        prompt="Tmux config file",
        default="~/.tmux.conf",
    ),
    description="📺 Setup `tmux`.",
    group=setup_group,
    alias="tmux",
)
def setup_tmux(ctx: AnyContext):
    tmux_config = read_file(os.path.join(os.path.dirname(__file__), "tmux_config.sh"))
    tmux_config_file = os.path.expanduser(ctx.input["tmux-config"])
    # Make sure config file exists
    if not os.path.isfile(tmux_config_file):
        write_file(tmux_config_file, "")
    content = read_file(tmux_config_file)
    if tmux_config in content:
        return
    # Write config
    write_file(tmux_config_file, [content, tmux_config, ""])
    ctx.print("Setup complete, restart your terminal to continue")


install_tmux >> setup_tmux
